import json
import os
import subprocess
import yaml
from os import path
from scyllaso.util import log, call, LogLevel


def apply(terraform_plan, workspace=None, options=""):
    log(f'Using terraform_plan [{terraform_plan}]')

    if not path.isdir(terraform_plan):
        log(f"Could not find directory [{terraform_plan}]", log_level=LogLevel.error)
        exit(1)

    cmd = f'terraform -chdir={terraform_plan} init'
    log(cmd)
    exitcode = call(cmd)
    if exitcode != 0:
        raise Exception(f'Failed terraform init, plan [{terraform_plan}], exitcode={exitcode} command=[{cmd}])')

    if workspace:
        cmd = f'terraform -chdir={terraform_plan} workspace select {workspace} || terraform -chdir={terraform_plan} workspace new {workspace}'
        log(cmd)
        exitcode = call(cmd, shell=True, split=False)
        if exitcode != 0:
            raise Exception(f'Failed terraform workspace, plan [{terraform_plan}], exitcode={exitcode} command=[{cmd}])')
        profile_file = os.path.join(os.path.abspath(os.getcwd()), workspace + ".yml")
        if path.isfile(profile_file):
            options = f'{options} -var="yaml_configuration_path={profile_file}"'

    options = f'{options} -auto-approve'

    cmd = f'terraform -chdir={terraform_plan} apply {options}'
    log(cmd)
    exitcode = call(cmd, shell=True, split=False)
    if exitcode != 0:
        raise Exception(f'Failed terraform apply, plan [{terraform_plan}], exitcode={exitcode} command=[{cmd}])')

    create_environment_yaml(terraform_plan, workspace)


def destroy(terraform_plan, workspace=None, options=""):
    log(f'Using terraform_plan [{terraform_plan}]')

    if not path.isdir(terraform_plan):
        log(f"Could not find directory [{terraform_plan}]", log_level=LogLevel.error)
        exit(1)

    if workspace:
        cmd = f'terraform -chdir={terraform_plan} workspace select {workspace}'
        log(cmd)
        exitcode = call(cmd)
        if exitcode != 0:
            raise Exception(f'Failed terraform workspace, plan [{terraform_plan}], exitcode={exitcode} command=[{cmd}])')
        profile_file = os.path.join(os.path.abspath(os.getcwd()), workspace + ".yml")
        if path.isfile(profile_file):
            options = f'{options} -var="yaml_configuration_path={profile_file}"'

    options = f'{options} -auto-approve'
    cmd = f'terraform -chdir={terraform_plan} destroy {options}'
    log(cmd)
    exitcode = call(cmd, shell=True, split=False)
    if exitcode != 0:
        raise Exception(f'Failed terraform destroy, plan [{terraform_plan}], exitcode={exitcode} command=[{cmd}])')

    filename = filename_environment_yml(workspace)
    if os.path.exists(filename):
        os.remove(filename)

    if workspace:
        cmd = f'terraform -chdir={terraform_plan} workspace select default && terraform -chdir={terraform_plan} workspace delete {workspace}'
        print(cmd)
        exitcode = call(cmd, shell=True, split=False)
        if exitcode != 0:
            raise Exception(f'Failed terraform workspace, plan [{terraform_plan}], exitcode={exitcode} command=[{cmd}])')


# Extracts all the output from a sso_terraform directory and places it into an environment.yaml
def create_environment_yaml(dir, workspace):
    output_text = subprocess.check_output(f'terraform -chdir={dir} output -json', shell=True, text=True)
    output = json.loads(output_text)

    environment = {}
    for key, value in output.items():
        environment[key] = output[key]['value']

    filename = filename_environment_yml(workspace)
    with open(filename, 'w') as environment_yml:
        yaml.dump(environment, environment_yml)


def filename_environment_yml(workspace):
    if workspace:
        return f'environment_{workspace}.yml'
    else:
        return 'environment.yml'



from scyllaso import common
from scyllaso.ssh import PSSH


def cli():
    props = common.load_yaml('properties.yml')
    env = common.load_yaml('environment.yml')
    pssh = PSSH(env['loadgenerator_public_ips'], props['load_generator_user'], props['ssh_options'])
    pssh.exec(f'killall -q -9 java')
    pssh.exec(f'killall -q -9 go/bin/scylla-bench')



from datetime import datetime
from time import sleep
from scyllaso.ssh import PSSH, SSH
from scyllaso.util import log, run_parallel, log_important
from scyllaso.cql import wait_for_cql_start


def clear_cluster(cluster_public_ips, cluster_user, ssh_options, duration_seconds=90):
    log("Shutting down cluster and removing all data")
    pssh = PSSH(cluster_public_ips, cluster_user, ssh_options)
    # pssh.exec("nodetool flush")
    log("Stopping scylla")
    pssh.exec("sudo systemctl stop scylla-server")
    log("Removing data dir")
    pssh.exec("sudo rm -fr /var/lib/scylla/data/*")
    log("Removing commit log")
    pssh.exec("sudo rm -fr /var/lib/scylla/commitlog/*")
    log("Starting scylla")
    pssh.exec("sudo systemctl start scylla-server")
    log(f"Waiting {duration_seconds} seconds")
    sleep(duration_seconds)
    log("Cluster cleared and restarted")


def restart_cluster(cluster_public_ips, cluster_user, ssh_options, duration_seconds=90):
    log("Restart cluster ")
    pssh = PSSH(cluster_public_ips, cluster_user, ssh_options)
    log("nodetool drain")
    pssh.exec("nodetool drain")
    log("sudo systemctl stop scylla-server")
    pssh.exec("sudo systemctl stop scylla-server")
    log("sudo systemctl start scylla-server")
    pssh.exec("sudo systemctl start scylla-server")
    log(f"Waiting {duration_seconds} seconds")
    sleep(duration_seconds)
    log("Cluster restarted")


def nodes_remove_data(cluster_user, ssh_options, *public_ips):
    log(f"Removing data from nodes {public_ips}")
    pssh = PSSH(public_ips, cluster_user, ssh_options)
    pssh.exec("sudo rm -fr /var/lib/scylla/data/*")
    pssh.exec("sudo rm -fr /var/lib/scylla/commitlog/*")
    log(f"Removing data from nodes {public_ips}: done")


def nodes_stop(cluster_user, ssh_options, *public_ips):
    log(f"Stopping nodes {public_ips}")
    pssh = PSSH(public_ips, cluster_user, ssh_options)
    pssh.exec("nodetool flush")
    pssh.exec("sudo systemctl stop scylla-server")
    log(f"Stopping nodes {public_ips}: done")


def nodes_start(cluster_user, ssh_options, *public_ips):
    log(f"Starting nodes {public_ips}")
    pssh = PSSH(public_ips, cluster_user, ssh_options)
    pssh.exec("sudo systemctl start scylla-server")
    log(f"Starting nodes {public_ips}: done")


# Assumes Scylla was started from official Scylla AMI.
class Scylla:

    def __init__(self, cluster_public_ips, cluster_private_ips, seed_private_ip, properties):
        self.properties = properties
        self.cluster_public_ips = cluster_public_ips
        self.cluster_private_ips = cluster_private_ips
        self.seed_private_ip = seed_private_ip
        self.ssh_user = properties['cluster_user']

    def __new_ssh(self, ip):
        return SSH(ip, self.ssh_user, self.properties['ssh_options'])

    def __install(self, ip):
        ssh = self.__new_ssh(ip)

        # Scylla AMI automatically performs setup
        # and then starts up. Each node is a separate 1-node cluster.
        # Here, we wait for this startup.
        wait_for_cql_start(ip)

        # Scylla started. Now we stop it and wipe
        # the data it generated.

        # FIXME - stop scylla-server more forcefully?
        ssh.exec("sudo systemctl stop scylla-server")
        ssh.exec("sudo rm -rf /var/lib/scylla/data/*")
        ssh.exec("sudo rm -rf /var/lib/scylla/commitlog/*")

        # Patch configuration files
        ssh.exec("sudo sed -i \"s/cluster_name:.*/cluster_name: cluster1/g\" /etc/scylla/scylla.yaml")
        ssh.exec(f'sudo sed -i \"s/seeds:.*/seeds: {self.seed_private_ip} /g\" /etc/scylla/scylla.yaml')
        ssh.exec("sudo sh -c \"echo 'compaction_static_shares: 100' >> /etc/scylla/scylla.yaml\"")
        ssh.exec("sudo sh -c \"echo 'compaction_enforce_min_threshold: true' >> /etc/scylla/scylla.yaml\"")

    def install(self):
        log_important("Installing Scylla: started")
        run_parallel(self.__install, [(ip,) for ip in self.cluster_public_ips])
        log_important("Installing Scylla: done")

    def append_configuration(self, configuration):
        print(f"Appending configuration to nodes {self.cluster_public_ips}: {configuration}")
        pssh = PSSH(self.cluster_public_ips, self.ssh_user, self.properties['ssh_options'])
        pssh.exec(f"sudo sh -c \"echo '{configuration}' >> /etc/scylla/scylla.yaml\"")

    def start(self):
        print(f"Starting Scylla nodes {self.cluster_public_ips}")
        for public_ip in self.cluster_public_ips:
            ssh = self.__new_ssh(public_ip)
            ssh.exec("sudo systemctl start scylla-server")
            wait_for_cql_start(public_ip)
            print("Node finished bootstrapping at:", datetime.now().strftime("%H:%M:%S"), public_ip)
        print(f"Starting Scylla nodes {self.cluster_public_ips}: done")

    def nodetool(self, command, load_index=None):
        if load_index is None:
            run_parallel(self.nodetool, [(command, i) for i in range(len(self.cluster_private_ips))])
        else:
            ssh = self.__new_ssh(self.cluster_public_ips[load_index])
            ssh.exec(f"nodetool {command}")

    def stop(self, load_index=None, erase_data=False):
        if load_index is None:
            print("Not implemented!")
        else:
            self.nodetool("drain", load_index=load_index)
            ssh = self.__new_ssh(self.cluster_public_ips[load_index])
            ssh.exec("sudo systemctl stop scylla-server")

            if erase_data:
                ssh.exec("sudo rm -rf /var/lib/scylla/data/*")
                ssh.exec("sudo rm -rf /var/lib/scylla/commitlog/*")



import yaml
import argparse
from scyllaso import terraform


def get_plan(args):
    terraform_plan = args.terraform_plan
    if not terraform_plan:
        with open('properties.yml') as f:
            properties = yaml.load(f, Loader=yaml.FullLoader)
        terraform_plan = properties.get('terraform_plan')
        if not terraform_plan:
            raise Exception("Could not find 'terraform_plan' in properties.yml")
    return terraform_plan


def provision():
    parser = argparse.ArgumentParser()
    parser.add_argument("terraform_plan", help="The terraform_plan to execute (directory).", nargs='?')
    parser.add_argument("--options", nargs='?', default="")
    parser.add_argument("--workspace", nargs='?')
    args = parser.parse_args()
    terraform.apply(get_plan(args), workspace=args.workspace, options=args.options)


def unprovision():
    parser = argparse.ArgumentParser()
    parser.add_argument("terraform_plan", help="The terraform_plan to execute (directory).", nargs='?')
    parser.add_argument("--options", nargs='?', default="")
    parser.add_argument("--workspace", nargs='?')
    args = parser.parse_args()
    terraform.destroy(get_plan(args), workspace=args.workspace, options=args.options)



import os
import selectors
import subprocess
import shlex
import time
from scyllaso.util import run_parallel, log_machine, log, LogLevel, WorkerThread


# Parallel SSH
class PSSH:

    def __init__(self,
                 ip_list,
                 user,
                 ssh_options,
                 use_control_socket=True,
                 silent_seconds=30,
                 log_ssh=False):
        self.ip_list = ip_list
        self.user = user
        self.ssh_options = ssh_options
        self.use_control_socket = use_control_socket
        self.silent_seconds = silent_seconds
        self.log_ssh = log_ssh

    def __new_ssh(self, ip):
        return SSH(ip,
                   self.user,
                   self.ssh_options,
                   use_control_socket=self.use_control_socket,
                   silent_seconds=self.silent_seconds,
                   log_ssh=self.log_ssh)

    def __exec(self, ip, cmd):
        self.__new_ssh(ip).exec(cmd)

    def exec(self, cmd):
        run_parallel(self.__exec, [(ip, cmd) for ip in self.ip_list])

    def async_exec(self, command):
        thread = WorkerThread(self.exec, (command))
        thread.start()
        return thread.future

    def __update(self, ip):
        self.__new_ssh(ip).update()

    def update(self):
        # todo: needs to be fixed; should be parallel, now sequential
        for ip in self.ip_list:
            ssh = SSH(ip, self.user, self.ssh_options)
            ssh.update()

        # can't get this to run correctly.
        # run_parallel(self.__update, [ip for ip in self.ip_list])

    def __install_one(self, ip, *packages):
        self.__new_ssh(ip).install_one(*packages)

    def install_one(self, *packages):
        run_parallel(self.__install_one, [(ip, *packages) for ip in self.ip_list])

    def __try_install(self, ip, *packages):
        self.__new_ssh(ip).try_install(*packages)

    def try_install(self, *packages):
        run_parallel(self.__try_install, [(ip, *packages) for ip in self.ip_list])

    def __install(self, ip, *packages):
        self.__new_ssh(ip).install(*packages)

    def install(self, *packages):
        run_parallel(self.__install, [(ip, *packages) for ip in self.ip_list])

    def __scp_from_remote(self, src, dst_dir, ip):
        self.__new_ssh(ip).scp_from_remote(src, os.path.join(dst_dir, ip))

    def scp_from_remote(self, src, dst_dir):
        run_parallel(self.__scp_from_remote, [(src, dst_dir, ip) for ip in self.ip_list])

    def __scp_to_remote(self, src, dst, ip):
        self.__new_ssh(ip).scp_to_remote(src, dst)

    def scp_to_remote(self, src, dst):
        run_parallel(self.__scp_to_remote, [(src, dst, ip) for ip in self.ip_list])

    def __set_governor(self, ip, governor):
        self.__new_ssh(ip).set_governor(governor)

    def set_governor(self, governor):
        run_parallel(self.__set_governor, [(ip, governor) for ip in self.ip_list])


class SSH:

    def __init__(self,
                 ip,
                 user,
                 ssh_options,
                 silent_seconds=30,
                 use_control_socket=True,
                 log_ssh=False):
        self.ip = ip
        self.user = user
        self.ssh_options = ssh_options
        self.silent_seconds = silent_seconds
        self.log_ssh = log_ssh
        self.prefix = "    " + f"[{self.ip}]".ljust(17, " ")
        if use_control_socket:
            self.control_socket_file = f"/tmp/{self.user}@{self.ip}.socket"
        else:
            self.control_socket_file = None

    def __wait_for_connect(self):
        if self.control_socket_file:
            if os.path.exists(self.control_socket_file):
                return
            cmd = f'ssh -M -S {self.control_socket_file} -o ControlPersist=5m {self.ssh_options} {self.user}@{self.ip} exit'
        else:
            cmd = f'ssh {self.ssh_options} {self.user}@{self.ip} exit'

        exitcode = None
        for i in range(1, 300):
            if i > self.silent_seconds:
                log(f"{self.prefix} Connect to {self.ip}")
                exitcode = subprocess.call(cmd, shell=True)
            else:
                exitcode = subprocess.call(cmd, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

            if exitcode == 0 or exitcode == 1:  # todo: we need to deal better with exit code
                self.wait_for_connect = False
                return
            time.sleep(1)

        raise Exception(f"Failed to connect to {self.ip}, exitcode={exitcode}")

    def __is_connected(self):
        return self.control_socket_file and os.path.exists(self.control_socket_file)

    def scp_from_remote(self, src, dst_dir):
        os.makedirs(dst_dir, exist_ok=True)
        cmd = f'scp {self.ssh_options} -r -q {self.user}@{self.ip}:{src} {dst_dir}'
        self.__scp(cmd)

    def scp_to_remote(self, src, dst):
        cmd = f'scp {self.ssh_options} -r -q {src} {self.user}@{self.ip}:{dst}'
        self.__scp(cmd)

    def __scp(self, cmd):
        self.__wait_for_connect()
        exitcode = subprocess.call(cmd, shell=True)
        # raise Exception(f"Failed to execute {cmd} after {self.max_attempts} attempts")

    def exec(self, command, ignore_errors=False):
        self.__wait_for_connect()

        socket = ""
        if self.__is_connected():
            socket = f"-S {self.control_socket_file}"

        cmd = f'ssh {socket} {self.ssh_options} {self.user}@{self.ip} \'{command}\''
        if self.log_ssh:
            log_machine(self.ip, cmd)

        process = subprocess.Popen(shlex.split(cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        sel = selectors.DefaultSelector()
        sel.register(process.stdout, selectors.EVENT_READ)
        sel.register(process.stderr, selectors.EVENT_READ)

        while True:
            for key, _ in sel.select():
                data = key.fileobj.read1().decode()
                if not data:
                    exitcode = process.poll()
                    if ignore_errors or exitcode == 0 or exitcode == 1:  # todo: we need to deal better with exit code
                        return
                    else:
                        raise Exception(f"Failed to execute {cmd}, exitcode={exitcode}")
                lines = data.splitlines()
                log_level = LogLevel.info if key.fileobj is process.stdout else LogLevel.warning
                for line in lines:
                    log_machine(self.ip, line, log_level)

    def async_exec(self, command):
        thread = WorkerThread(self.exec, (command))
        thread.start()
        return thread.future

    def update(self):
        log_machine(self.ip, f'Update: started')
        self.exec(
            f"""
            set -e
            if [ -f /tmp/update.called ] ; then
                # echo "Skipping update"
                exit 0
            fi

            if hash apt-get 2>/dev/null; then
                sudo apt-get -y -qq update
            elif hash yum 2>/dev/null; then
                sudo yum -y -q update
            else
                echo "Cannot update: yum/apt not found"
                exit 1
            fi

            touch /tmp/update.called
            """)
        log_machine(self.ip, f'Update: done')

    def install_one(self, *packages):
        self.exec(
            f"""
            set -e
            for package in {" ".join(packages)} 
            do                
                echo Trying package [$package]
                if hash apt-get 2>/dev/null ; then
                    if sudo apt show $package >/dev/null 2>&1; then
                        echo      [{self.ip}] Installing $package
                        sudo apt-get install -y -qq $package
                        exit 0
                    fi    
                elif hash yum 2>/dev/null; then
                    if sudo yum info $package >/dev/null 2>&1; then
                        echo      [{self.ip}] Installing $package
                        sudo yum -y -q install $package
                        exit 0
                    fi                        
                else
                    echo "     [{self.ip}] Cannot install $package: yum/apt not found"
                    exit 1
                fi
                    
                echo Not found $package                       
            done
            echo "Could not find any of the packages from {packages}"
            exit 1
            """)

    def try_install(self, *packages):
        self.install(*packages, ignore_errors=True)

    def install(self, *packages, ignore_errors=False):
        for package in packages:
            log_machine(self.ip, f'Install: {package}')
            self.exec(
                f"""
                set -e
                if hash apt-get 2>/dev/null; then
                    sudo apt-get install -y -qq {package}
                elif hash yum 2>/dev/null; then
                    sudo yum -y -q install {package}
                else
                    echo "Cannot install {package}: yum/apt not found"
                    exit 1
                fi
                """, ignore_errors=ignore_errors)

    def set_file_property(self, file_path, property, seperator, value):
        self.exec(f"""
            set -e
            sudo touch {file_path}
            if grep -q -E "^\\s*{property}\\s*{seperator}.*" {file_path}; then 
                sudo sed -i "s/^\\s*{property}\\s*{seperator}.*/{property + seperator + value}/g" {file_path}
            else
                sudo sh -c "'echo {property + seperator + value}' >> {file_path}"
            fi
        """)

    def set_governor(self, governor):
        log_machine(self.ip, f'Set governor [{governor}]')
        self.exec(f"""
            set -e
    
            if ! hash cpupower 2>/dev/null; then
                echo "Installing cpupower"
                if hash apt-get 2>/dev/null; then
                    sudo apt-get -y -qq update
                    version=$(uname -r)
                    sudo apt-get -y -qq install linux-tools-$version
                    echo "apt-get found"                           
                elif hash yum 2>/dev/null; then
                    sudo yum -y -q install kernel-tools                
                else
                    echo "Cannot install {governor}: yum/apt not found"                
                    exit 1
                fi
            fi
            echo "=========== frequency info before change ==========="
            sudo cpupower frequency-info
            echo "===================================================="
            frequencyinfo=$(sudo cpupower frequency-info)                  
            if [[ $frequencyinfo =~ "{governor}" ]]; then
                sudo cpupower frequency-set -g {governor}  
                echo "=========== frequency info after change ==========="
                sudo cpupower frequency-info
                echo "===================================================="                      
            else
                echo "Governor [{governor}] is not supported"
            fi             
        """)



import os
import time

from datetime import datetime
from scyllaso.ssh import SSH
from scyllaso.util import run_parallel, WorkerThread, log_important, log_machine, log

class ScyllaBench:

    def __init__(self, load_ips, properties, performance_governor=True):
        self.properties = properties
        self.load_ips = load_ips
        self.performance_governor = performance_governor

    def __new_ssh(self, ip):
        return SSH(ip, self.properties['load_generator_user'], self.properties['ssh_options'])

    def __install(self, ip):
        ssh = self.__new_ssh(ip)
        log_machine(ip, f'Installing scylla_bench: started')
        ssh.update()

        if self.performance_governor:
            ssh.set_governor("performance")

        ssh.install("golang")
        #ssh.exec("go get github.com/scylladb/scylla-bench")

        ssh.exec(f"""
                  git clone https://github.com/scylladb/scylla-bench
                  cd scylla-bench/
                  go install .
                  """)

        log_machine(ip, f'Installing scylla_bench: done')

    def install(self):
        log_important("Installing scylla_bench: started")
        run_parallel(self.__install, [(ip,) for ip in self.load_ips])
        log_important("Installing scylla_bench: done")

    def __stress(self, ip, cmd):
        full_cmd = f'go/bin/scylla-bench {cmd}'
            
        dt=datetime.now().strftime("%d-%m-%Y_%H-%M-%S")
        full_cmd = full_cmd + f" 2>&1 | tee -a scylla-bench-{dt}.log"    
        log(full_cmd)
        self.__new_ssh(ip).exec(full_cmd)
      
    def stress(self, command, load_index=None):
        if load_index is None:
            log_important("scylla-bench: started")
            run_parallel(self.__stress, [(ip, command) for ip in self.load_ips])
            log_important("scylla-bench: done")
        else:
            log("using load_index " + str(load_index))
            self.__stress(self.load_ips[load_index], command)

    def async_stress(self, command, load_index=None):
        thread = WorkerThread(self.stress, (command, load_index))
        thread.start()
        return thread.future

    def insert(self, 
               partition_count, 
               nodes, 
               partition_offset=0,
               concurrency=64, 
               clustering_row_count=1, 
               extra_args=""):
        log_important(f"Inserting {partition_count} partitions")
        start_seconds = time.time()

        # todo: there could be some loss if there is a reaminer.
        pc_per_lg = partition_count // len(self.load_ips)

        cmd_list = []
        for i in range(0, len(self.load_ips)):
            cmd = f"""-workload sequential \
                      -clustering-row-count {clustering_row_count} \
                      -mode write \
                      -partition-count {pc_per_lg} \
                      -partition-offset {partition_offset} \
                      -nodes {nodes} \
                      -concurrency {concurrency} \ 
                      {extra_args}"""
            # clean the string up.
            cmd = " ".join(cmd.split())
            cmd_list.append(cmd)
            partition_offset = partition_offset + pc_per_lg

        futures = []
        for i in range(0, len(self.load_ips)):
            f = self.async_stress(cmd_list[i], load_index=i)
            futures.append(f)
            if i == 0:
                # first one is given some extra time to set up the tables and all that.
                time.sleep(10)

        for f in futures:
            f.join()

        duration_seconds = time.time() - start_seconds
        log(f"Duration : {duration_seconds} seconds")
        log(f"Insertion rate: {partition_count // duration_seconds} items/second")
        log_important(f"Inserting {partition_count} partitions: done")

    def __ssh(self, ip, command):
        self.__new_ssh(ip).exec(command)

    def ssh(self, command):
        run_parallel(self.__ssh, [(ip, command) for ip in self.load_ips])

    def __upload(self, ip, file):
        self.__new_ssh(ip).scp_to_remote(file, "")

    def upload(self, file):
        log_important(f"Upload: started")
        run_parallel(self.__upload, [(ip, file) for ip in self.load_ips])
        log_important(f"Upload: done")

    def __collect(self, ip, dir):
        dest_dir = os.path.join(dir, ip)
        os.makedirs(dest_dir, exist_ok=True)
        log_machine(ip, f'Collecting to [{dest_dir}]')
        ssh = self.__new_ssh(ip)
        ssh.scp_from_remote(f'*.log', dest_dir)
        ssh.exec(f'rm -fr *.log')
        log_machine(ip, f'Collecting to [{dest_dir}] done')

    def collect_results(self, dir):
        """
        Parameters
        ----------
        dir: str
            The download directory.
        """

        log_important(f"Collecting results: started")
        run_parallel(self.__collect, [(ip, dir) for ip in self.load_ips])
        log_important(f"Collecting results: done")
        log(f"Results can be found in [{dir}]")
     
    def __prepare(self, ip):
        log_machine(ip, f'Preparing: started')
        ssh = self.__new_ssh(ip)
        # we need to make sure that the no old load generator is still running.
        ssh.exec(f'killall -q -9 go/bin/scylla-bench')
        log_machine(ip, f'Preparing: done')

    def prepare(self):
        log_important(f"Preparing load generator: started")
        run_parallel(self.__prepare, [(ip,) for ip in self.load_ips])
        log_important(f"Preparing load generator: done")



#!/bin/python3

import sys
from scyllaso import common
from scyllaso.cs import CassandraStress
from scyllaso.common import Iteration
from scyllaso.scylla import Scylla
from scyllaso.cassandra import Cassandra
from datetime import datetime

print("Test started at:", datetime.now().strftime("%H:%M:%S"))

if len(sys.argv) < 2:
    raise Exception("Usage: ./benchmark_latency_throughput.py [PROFILE_NAME]")

# Load properties
profile_name = sys.argv[1]
props = common.load_yaml(f'{profile_name}.yml')
env = common.load_yaml(f'environment_{profile_name}.yml')
cluster_private_ips = env['cluster_private_ips']
cluster_string = ",".join(cluster_private_ips)
cluster_public_ips = env['cluster_public_ips']
loadgenerator_public_ips = env['loadgenerator_public_ips']
loadgenerator_count = len(loadgenerator_public_ips)

# Run parameters

# Row size of default cassandra-stress workload.
# Measured experimentally.
ROW_SIZE_BYTES = 210 * 1024 * 1024 * 1024 / 720_000_000

TARGET_DATASET_SIZE = props['target_dataset_size_gb'] * 1024 * 1024 * 1024

REPLICATION_FACTOR = 1
COMPACTION_STRATEGY = props['compaction_strategy']
ROW_COUNT = int(TARGET_DATASET_SIZE / ROW_SIZE_BYTES / REPLICATION_FACTOR)

# Start Scylla/Cassandra nodes
if props['cluster_type'] == 'scylla':
    cluster = Scylla(env['cluster_public_ips'], env['cluster_private_ips'], env['cluster_private_ips'][0], props)
    cluster.install()
    cluster.start()
else:
    cluster = Cassandra(env['cluster_public_ips'], env['cluster_private_ips'], env['cluster_private_ips'][0], props)
    cluster.install()
    if "cassandra_extra_env_opts" in props:
        cluster.append_env_configuration(props["cassandra_extra_env_opts"])
    cluster.start()

print("Nodes started at:", datetime.now().strftime("%H:%M:%S"))

cs = CassandraStress(env['loadgenerator_public_ips'], props)
cs.install()
cs.prepare()

print("Loading started at:", datetime.now().strftime("%H:%M:%S"))

THROTTLE = props["loading_total_throttle"] // loadgenerator_count
cs.stress_seq_range(ROW_COUNT, 'write cl=QUORUM', f'-schema "replication(strategy=SimpleStrategy,replication_factor={REPLICATION_FACTOR})" "compaction(strategy={COMPACTION_STRATEGY})" -log hdrfile=profile.hdr -graph file=report.html title=benchmark revision=benchmark-0 -mode native cql3 maxPending=1024 -rate "threads=300 throttle={THROTTLE}/s" -node {cluster_string}')

cluster.nodetool("flush")

confirm = input("Has compaction finished? Input 'yes':")
while confirm != 'yes':
    confirm = input("Has compaction finished? Input 'yes':")

print("Run started at:", datetime.now().strftime("%H:%M:%S"))

# Background load
BACKGROUND_LOAD_OPS = props["background_total_load_ops"] // loadgenerator_count
background_load = cs.loop_stress(f'mixed ratio\\(write=1,read=1\\) duration=5m cl=QUORUM -pop dist=UNIFORM\\(1..{ROW_COUNT}\\) -log hdrfile=profile.hdr -graph file=report.html title=benchmark revision=benchmark-0 -mode native cql3 -rate "threads=700 fixed={BACKGROUND_LOAD_OPS}/s" -node {cluster_string}')

iteration = Iteration(f'{profile_name}/compact', ignore_git=True)

compact_start = datetime.now()

for i in range(len(cluster_public_ips)):
    print("Compacting node", i, "started at:", datetime.now().strftime("%H:%M:%S"))
    cluster.nodetool("compact", i)
    print("Compacting node", i, "ended at:", datetime.now().strftime("%H:%M:%S"))

compact_end = datetime.now()

with open(f'{iteration.dir}/result.txt', 'a') as writer:
    writer.write(f'Major compaction on all nodes took (s): {(compact_end - compact_start).total_seconds()}\n')

print("Run ended at:", datetime.now().strftime("%H:%M:%S"))

background_load.request_stop()
background_load.join()
print("Background load ended:", datetime.now().strftime("%H:%M:%S"))



from scyllaso.ssh import PSSH,SSH
from scyllaso.util import log_important, log

class Perf:

    def __init__(self, ip_list, user, ssh_options):
        log(ip_list)
        self.updated = False
        self.ip_list = ip_list
        self.user = user
        self.ssh_options = ssh_options

    def pssh(self):
        return PSSH(self.ip_list, self.user, self.ssh_options)

    def install(self):
        #pssh = self.pssh()
        #pssh.exec
        self.install_perf()
        self.install_flamegraph()
        self.install_debuginfo()
        #pssf.exec("touch /tmp/perf_installed")

    def install_debuginfo(self):
        pssh = self.pssh()
        
        if not self.updated:
            pssh.update()
            self.updated = True
        
        log_important("Installing debuginfo: started")
        pssh.try_install("scylla_debuginfo")
        log_important("Installing debuginfo: done")

    def install_perf(self):
        log_important("Perf install: started")
        pssh = self.pssh()
        
        if not self.updated:
            pssh.update()
            self.updated = True
       
        # This part sucks.. Should no be a dependency on a particular version 
        pssh.install_one("perf", "linux-tools-5.4.0-1035-aws")        
        log_important("Perf install: done")

    def install_flamegraph(self):
        log_important("Perf install flamegraph: started")
        pssh = self.pssh()

        if not self.updated:
            pssh.update()
            self.updated = True

        pssh.install("git")
        # needed for addr2line
        pssh.install("binutils")
        pssh.exec(f"""
                cd /tmp
                if [ ! -d FlameGraph ]; then
                    echo "cloning flamegraph"
                    git clone https://github.com/brendangregg/FlameGraph
                fi
                """)
        log_important("Perf install flamegraph: done")

    def flamegraph_cpu(self, cpu, dir, duration_seconds=60, args = "--call-graph lbr -F99", output="flamegraph"):
        data_file = f"{output}.data"
        flamegraph_file = f"{output}.svg"
        cmd = f"--output {data_file} --cpu {cpu} {args} sleep {duration_seconds}"
        self.record(cmd)
        self.collect_flamegraph(dir, data_file, flamegraph_file)

    def list(self):
        self.exec("sudo perf list -v")

    def record(self, command):
        cmd = f"sudo perf record {command}"
        self.exec(cmd)

    def script(self, command):
        cmd = f"sudo perf script {command}"
        self.exec(cmd)

    def exec(self, command):
        log_important(f"Perf: started")
        log(command)
        ssh = SSH(self.ip_list[0], self.user, self.ssh_options)
        ssh.exec(f"""
                cd /tmp
                {command}
                """)
        log_important(f"Perf: done")

    def collect_flamegraph(self, dir, data_file = "perf.data", flamegraph_file = "flamegraph.svg"):
        log_important(f"Perf collecting flamegraph: started")
        ssh = SSH(self.ip_list[0], self.user, self.ssh_options)
        # --no-online
        ssh.exec(f"""
                cd /tmp
                sudo perf script -i {data_file} | FlameGraph/stackcollapse-perf.pl | FlameGraph/flamegraph.pl --hash > {flamegraph_file}
                """)
        ssh.scp_from_remote(f"/tmp/{flamegraph_file}", dir)
        ssh.exec(f"rm /tmp/{flamegraph_file}")
        log_important(f"Perf collecting flamegraph: done")



import os
import yaml
import subprocess
import time
from datetime import datetime
from scyllaso.ssh import SSH
from scyllaso.util import run_parallel


def load_yaml(path):
    with open(path) as f:
        return yaml.load(f, Loader=yaml.FullLoader)


class Iteration:

    # The purpose of the experimental flag is to modify the name of the trial directory to easily
    # weed out the experimental runs from non experimental ones. Otherwise it is easy to run in a 
    # messed up history where it isn't clear what is experimental and what isn't.
    def __init__(self, trial_name, description=None, experimental=False, ignore_git=True):
        self.trials_dir_name = "trials"
        self.trials_dir = os.path.join(os.getcwd(), self.trials_dir_name)
        self.trial_name = trial_name
        self.trial_dir = os.path.join(self.trials_dir, trial_name)

        while True:
            self.name = datetime.now().strftime("%d-%m-%Y_%H-%M-%S")
            if experimental:
                self.name = self.name + "_experimental"
            self.dir = os.path.join(self.trial_dir, self.name)
            if not os.path.exists(self.dir):
                os.makedirs(self.dir)
                break
            time.sleep(0.5)
           
        if description:
            desc_file = os.path.join(self.dir, "description.txt")
            with open(desc_file, "w") as text_file:
                print(description, file=text_file)

        if not experimental:
            latest_dir = os.path.join(self.trial_dir, "latest")
            if os.path.isdir(latest_dir):
                os.unlink(latest_dir)
            elif os.path.islink(latest_dir):
                # it is a broken sym link, so lets remove it.
                os.remove(latest_dir)
            os.symlink(self.dir, latest_dir, target_is_directory=True)

        if not ignore_git:    
            exitcode = subprocess.call("git status", shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            if exitcode == 0:
                if not experimental:
                    d = subprocess.check_output(" [[ -z $(git status -s) ]] || echo 'dirty'", shell=True).decode()
                    if d.startswith("dirty"):
                        print("The current working directory is dirty, so can't store the git HEAD in the iteration directory.")
                        exit(1)
            
                    output = subprocess.check_output("git log --pretty=format:'%h' -n 1", shell=True).decode()
                    git_file = os.path.join(self.dir, "HEAD")
                    with open(git_file, "w") as git_file:
                        print(output, file=git_file)

        print(f'Using iteration directory [{self.dir}]')
    

def __collect_ec2_metadata(ip, ssh_user, ssh_options, dir):
    dest_dir = os.path.join(dir, ip)
    os.makedirs(dest_dir, exist_ok=True)

    ssh = SSH(ip, ssh_user, ssh_options)
    ssh.update()
    ssh.install("curl")
    ssh.exec("curl http://169.254.169.254/latest/dynamic/instance-identity/document > metadata.txt")
    ssh.scp_from_remote("metadata.txt", dest_dir)


def collect_ec2_metadata(ips, ssh_user, ssh_options, dir):
    run_parallel(__collect_ec2_metadata, [(ip, ssh_user, ssh_options, dir) for ip in ips])



#!/bin/python3

import sys
from scyllaso import common
from scyllaso.common import Iteration
from datetime import datetime
from scyllaso import prometheus

print("Test started at:", datetime.now().strftime("%H:%M:%S"))

if len(sys.argv) < 2:
    raise Exception("Usage: ./download_prometheus.py [PROFILE_NAME]")

profile_name = sys.argv[1]

# Load properties
props = common.load_yaml(f'{profile_name}.yml')
env = common.load_yaml(f'environment_{profile_name}.yml')

iteration = Iteration(f'{profile_name}/prometheus-dump', ignore_git=True)

prometheus.download(env, props, iteration)



from time import sleep
from datetime import datetime, timedelta
import socket

def wait_for_port_start(node_ip, port, port_name, wait_reason='', timeout=900, connect_timeout=10, max_tries_per_second=2):
    print(f'    [{node_ip}] Waiting for {port_name} port to start{wait_reason}. This could take a while.')

    backoff_interval = 1.0 / max_tries_per_second
    timeout_point = datetime.now() + timedelta(seconds=timeout)

    feedback_interval = 20
    print_feedback_point = datetime.now() + timedelta(seconds=feedback_interval)

    while datetime.now() < timeout_point:
        # Make a probe connection.
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock: 
            sock.settimeout(connect_timeout)
            try:
                sock.connect((node_ip, port))
            except:
                # There was a problem connecting to the port.
                sleep(backoff_interval)
                if datetime.now() > print_feedback_point:
                    print_feedback_point = datetime.now() + timedelta(seconds=feedback_interval)
                    print(f'    [{node_ip}] Still waiting for {port_name} port to start...')
            else:
                print(f'    [{node_ip}] Successfully connected to {port_name} port.')
                return

    raise Exception(f'Waiting for {port_name} to start timed out after {timeout} seconds for node: {node_ip}.')

def wait_for_cql_start(node_ip, timeout=900, connect_timeout=10, max_tries_per_second=2):
    wait_for_port_start(node_ip, 9042, 'CQL', ' (meaning node bootstrap finished)', timeout=timeout, 
        connect_timeout=connect_timeout, max_tries_per_second=max_tries_per_second)


import argparse


def cli():
    parser = argparse.ArgumentParser()
    parser.add_argument("nr_cpus", help="Number of cpus", nargs=1)
    parser.add_argument("cpus", help="The cpus", nargs='*')

    args = parser.parse_args()

    nr_cpus = int(args.nr_cpus[0])

    cpu_bit_list = []
    for i in range(0, nr_cpus):
        cpu_bit_list.append('0')

    for cpu_string in args.cpus:
        cpu = int(cpu_string)
        cpu_bit_list[cpu] = '1'

    print(to_cpu_set(cpu_bit_list))
    print("irq_cpu_mask: " + to_irq_cpu_mask(cpu_bit_list))


def to_cpu_set(cpu_bit_list):
    cpu_set = "CPUSET=\"--cpuset "
    first = True
    for cpu in range(0, len(cpu_bit_list)):
        if cpu_bit_list[cpu] == '0':
            if first:
                first = False
            else:
                cpu_set = cpu_set + ","
            cpu_set = cpu_set + str(cpu)
    cpu_set = cpu_set + '"'
    return cpu_set


def to_irq_cpu_mask(cpu_bit_list):
    cpu_bit_list.reverse()

    s = "".join(cpu_bit_list)

    a = hex(int(s, 2))

    # remove the first 2 chars (0x)
    a = a[2:]

    # prefix with zero's
    a = a.zfill(24)

    r = ""
    for i in range(0, 24):
        if i % 8 == 0:
            if i > 0:
                r += ","
            r += "0x"
        r += a[i]
    return r



import os
import shutil
import getpass
import argparse
import pkg_resources
from scyllaso.bin import make_key


def cli():
    parser = argparse.ArgumentParser()
    parser.add_argument("name",
                        help="The name of the benchmark.")
    parser.add_argument("--template",
                        help="The name of the benchmark template.", default="default")
    parser.add_argument("--resourceid",
                        help="An extra id to make resources unique. By default the username is used.", nargs='?')
    args = parser.parse_args()

    name = args.name
    if args.resourceid:
        resourceid = args.resourceid
    else:
        resourceid = getpass.getuser()

    cwd = os.getcwd()
    target_dir = os.path.join(cwd, name)

    if os.path.exists(target_dir):
        print(f"Can't generate benchmark: dir/file [{target_dir}] already exists")
        exit(1)

    copy_template(target_dir, args.template)

    os.chdir(target_dir)
    make_key.cli()
    process_templates(target_dir, resourceid)
    os.chdir(cwd)


def copy_template(target_dir, template):
    module_dir = os.path.dirname(pkg_resources.resource_filename('scyllaso', '__init__.py'))
    benchmarks_dir = os.path.join(module_dir, "benchmarks")
    template_dir = os.path.join(benchmarks_dir, template)
    if not os.path.exists(template_dir):
        print(f"Template directory [{template_dir}] does not exist.")
        exit(1)

    shutil.copytree(template_dir, target_dir)

    python_cache = os.path.join(target_dir, "__pycache__")
    if os.path.exists(python_cache):
        shutil.rmtree(python_cache)


def process_templates(target_dir, resourceid):
    for subdir, dirs, files in os.walk(target_dir):
        for filename in files:
            filepath = subdir + os.sep + filename
            if os.access(filepath, os.W_OK):
                f = open(filepath, 'r')
                old_text = f.read()
                f.close()

                new_text = old_text.replace("<resourceid>", resourceid)
                f = open(filepath, 'w')
                f.write(new_text)
                f.close()



from scyllaso.ssh import PSSH


class RAID:
    def __init__(self, public_ips, user, device_name_wildcard, raid_device_name, level, properties):
        self.level = level
        self.public_ips = public_ips
        self.user = user
        self.device_name_wildcard = device_name_wildcard
        self.raid_device_name = raid_device_name
        self.properties = properties

    def install(self):
        ips = ','.join(self.public_ips)
        print(f'    [{ips}] raid: starting creating RAID')
        pssh = PSSH(self.public_ips, self.user, self.properties['ssh_options'])
        pssh.exec(f"""
            if [[ ! -b /dev/md/{self.raid_device_name} ]]; then
                sudo mdadm --create --verbose /dev/md/{self.raid_device_name} --chunk=256 --metadata=1.2 --level={self.level} --force --raid-devices=$(ls {self.device_name_wildcard} | wc -l) {self.device_name_wildcard}
                
                # /dev/md/raid_device_name maps to /dev/md[0-9]+
                MD_NAME=$(basename $(readlink /dev/md/{self.raid_device_name}))

                # Tuning
                sudo sh -c "echo 1 > /sys/block/$MD_NAME/queue/nomerges"
                sudo sh -c "echo 8 > /sys/block/$MD_NAME/queue/read_ahead_kb"
                sudo sh -c "echo none > /sys/block/$MD_NAME/queue/scheduler"

                sudo mkfs.xfs -f /dev/$MD_NAME
                mkdir {self.raid_device_name}
                sudo mount /dev/$MD_NAME {self.raid_device_name}
                sudo chown $(id -u) {self.raid_device_name}
            fi
        """)
        print(f'    [{ips}] raid: finished creating RAID')



#!/bin/python3

import sys
import os
import time

from scyllaso import common
from scyllaso.cs import CassandraStress
from scyllaso.common import Iteration
from scyllaso.scylla import Scylla
from scyllaso.hdr import parse_profile_summary_file
from scyllaso.cassandra import Cassandra
from datetime import datetime

print("Test started at:", datetime.now().strftime("%H:%M:%S"))

if len(sys.argv) < 2:
    raise Exception("Usage: ./benchmark_latency_throughput.py [PROFILE_NAME]")

# Load properties
profile_name = sys.argv[1]
props = common.load_yaml(f'{profile_name}.yml')
env = common.load_yaml(f'environment_{profile_name}.yml')
cluster_private_ips = env['cluster_private_ips']
cluster_string = ",".join(cluster_private_ips)
cluster_public_ips = env['cluster_public_ips']
loadgenerator_public_ips = env['loadgenerator_public_ips']
loadgenerator_count = len(loadgenerator_public_ips)

# Run parameters

# Row size of default cassandra-stress workload.
# Measured experimentally.
ROW_SIZE_BYTES = 210 * 1024 * 1024 * 1024 / 720_000_000

# 1TB per node
#TARGET_DATASET_SIZE = len(cluster_private_ips) * 1024 * 1024 * 1024 * 1024
TARGET_DATASET_SIZE = props['target_dataset_size_gb'] * 1024 * 1024 * 1024

REPLICATION_FACTOR = 3
COMPACTION_STRATEGY = props['compaction_strategy']
ROW_COUNT = int(TARGET_DATASET_SIZE / ROW_SIZE_BYTES / REPLICATION_FACTOR)

START_RATE     = props['start_rate'] #10000
RATE_INCREMENT = props['rate_increment'] #10000
DURATION_MINUTES = props['duration_minutes']

MAX_90_PERCENTILE_LATENCY = 1000.0

WRITE_COUNT = props['write_count']
READ_COUNT = props['read_count']

# Start Scylla/Cassandra nodes
if props['cluster_type'] == 'scylla':
    cluster = Scylla(env['cluster_public_ips'], env['cluster_private_ips'], env['cluster_private_ips'][0], props)
    cluster.install()
    cluster.start()
else:
    cluster = Cassandra(env['cluster_public_ips'], env['cluster_private_ips'], env['cluster_private_ips'][0], props)
    cluster.install()
    if "cassandra_extra_env_opts" in props:
        cluster.append_env_configuration(props["cassandra_extra_env_opts"])
    cluster.start()

print("Nodes started at:", datetime.now().strftime("%H:%M:%S"))

cs = CassandraStress(env['loadgenerator_public_ips'], props)
cs.install()
cs.prepare()

print("Loading started at:", datetime.now().strftime("%H:%M:%S"))

THROTTLE = (100000 // loadgenerator_count) if props['cluster_type'] == 'scylla' else (56000 // loadgenerator_count)

cs.stress_seq_range(ROW_COUNT, 'write cl=QUORUM', f'-schema "replication(strategy=SimpleStrategy,replication_factor={REPLICATION_FACTOR})" "compaction(strategy={COMPACTION_STRATEGY})" -log hdrfile=profile.hdr -graph file=report.html title=benchmark revision=benchmark-0 -mode native cql3 -rate "threads=700 throttle={THROTTLE}/s" -node {cluster_string}')

cluster.nodetool("flush")

confirm = input("Has compaction finished? Input 'yes':")
while confirm != 'yes':
    confirm = input("Has compaction finished? Input 'yes':")    

print("Run started at:", datetime.now().strftime("%H:%M:%S"))

rate = START_RATE

while True:
    print("Run iteration started at:", datetime.now().strftime("%H:%M:%S"))

    iteration = Iteration(f'{profile_name}/cassandra-stress-{rate}', ignore_git=True)

    cs.stress(f'mixed ratio\\(write={WRITE_COUNT},read={READ_COUNT}\\) duration={DURATION_MINUTES}m cl=QUORUM -pop dist=UNIFORM\\(1..{ROW_COUNT}\\) -log hdrfile=profile.hdr -graph file=report.html title=benchmark revision=benchmark-0 -mode native cql3 -rate "threads=500 fixed={rate // loadgenerator_count}/s" -node {cluster_string}')

    cs.collect_results(iteration.dir)

    if WRITE_COUNT > 0:
        write_profile_summary = parse_profile_summary_file(f'{iteration.dir}/profile-summary.txt', 'WRITE')
        print('WRITE_PROFILE', write_profile_summary)

    if READ_COUNT > 0:
        read_profile_summary = parse_profile_summary_file(f'{iteration.dir}/profile-summary.txt', 'READ')
        print('READ_PROFILE', read_profile_summary)

    with open(f'{iteration.dir}/parsed_profile_summary_file.txt', 'a') as writer:
        if WRITE_COUNT > 0:
            writer.write(f'WRITE_PROFILE: {write_profile_summary}\n')
        if READ_COUNT > 0:
            writer.write(f'READ_PROFILE: {read_profile_summary}\n')

    if WRITE_COUNT > 0 and write_profile_summary.p90_latency_ms > MAX_90_PERCENTILE_LATENCY:
        break
    if READ_COUNT > 0 and read_profile_summary.p90_latency_ms > MAX_90_PERCENTILE_LATENCY:
        break

    rate += RATE_INCREMENT

print("Run ended at:", datetime.now().strftime("%H:%M:%S"))



import subprocess
from os import path


def cli():
    if path.exists("key"):
        print("Key already found, skipping generation")
        return

    exitcode = subprocess.call(f"""ssh-keygen -P "" -m PEM -f key""", shell=True)
    if exitcode != 0:
        raise Exception(f'Failed to generate key')

    subprocess.call(f"""chmod 400 key""", shell=True)
    subprocess.call(f"""chmod 400 key""", shell=True)



#!/bin/python3

import sys
from scyllaso import common
from scyllaso.cs import CassandraStress
from scyllaso.common import Iteration
from scyllaso.scylla import Scylla
from scyllaso.cassandra import Cassandra
from datetime import datetime

print("Test started at:", datetime.now().strftime("%H:%M:%S"))

if len(sys.argv) < 2:
    raise Exception("Usage: ./benchmark_new_node.py [PROFILE_NAME]")

profile_name = sys.argv[1]

# Load properties
props = common.load_yaml(f'{profile_name}.yml')
env = common.load_yaml(f'environment_{profile_name}.yml')

start_count = props['start_count']
new_count = len(env['cluster_private_ips']) - start_count

cluster_private_ips = env['cluster_private_ips'][:start_count]
cluster_string = ",".join(cluster_private_ips[:start_count])
new_node_private_ips = env['cluster_private_ips'][start_count:]

cluster_public_ips = env['cluster_public_ips'][:start_count]
new_node_public_ips = env['cluster_public_ips'][start_count:]

all_public_ips = env['cluster_public_ips']
all_private_ips = env['cluster_private_ips']

loadgenerator_public_ips = env['loadgenerator_public_ips']
loadgenerator_count = len(loadgenerator_public_ips)

# Run parameters

# Row size of default cassandra-stress workload.
# Measured experimentally.
ROW_SIZE_BYTES = 210 * 1024 * 1024 * 1024 / 720_000_000

# 1TB per node
# TARGET_DATASET_SIZE = len(cluster_private_ips) * 1024 * 1024 * 1024 * 1024
TARGET_DATASET_SIZE = props['target_dataset_size_gb'] * 1024 * 1024 * 1024

REPLICATION_FACTOR = 3
COMPACTION_STRATEGY = props['compaction_strategy']
ROW_COUNT = int(TARGET_DATASET_SIZE / ROW_SIZE_BYTES / REPLICATION_FACTOR)

BACKGROUND_LOAD_OPS = 25000

# Start Scylla/Cassandra nodes (except ones to be started later)
if props['cluster_type'] == 'scylla':
    cluster = Scylla(all_public_ips, all_private_ips, all_private_ips[0], props)
    cluster.install()
    cluster = Scylla(cluster_public_ips, cluster_private_ips, cluster_private_ips[0], props)
    cluster.start()
else:
    cluster = Cassandra(all_public_ips, all_private_ips, all_private_ips[0], props)
    cluster.install()
    cluster = Cassandra(cluster_public_ips, cluster_private_ips, cluster_private_ips[0], props)
    cluster.start()

print("Nodes started at:", datetime.now().strftime("%H:%M:%S"))

# Setup cassandra stress
cs = CassandraStress(env['loadgenerator_public_ips'], props)
cs.install()
cs.prepare()

print("Loading started at:", datetime.now().strftime("%H:%M:%S"))

THROTTLE = (100000 * 5 // loadgenerator_count) if props['cluster_type'] == 'scylla' else (56000 * 9 // loadgenerator_count)

cs.stress_seq_range(ROW_COUNT, 'write cl=QUORUM', f'-schema "replication(strategy=SimpleStrategy,replication_factor={REPLICATION_FACTOR})" "compaction(strategy={COMPACTION_STRATEGY})" -log hdrfile=profile.hdr -graph file=report.html title=benchmark revision=benchmark-0 -mode native cql3 maxPending=1024 -rate "threads=700 throttle={THROTTLE}/s" -node {cluster_string}')

cluster.nodetool("flush")

confirm = input("Has compaction finished? Input 'yes':")
while confirm != 'yes':
    confirm = input("Has compaction finished? Input 'yes':")  

print("Run started at:", datetime.now().strftime("%H:%M:%S"))

# Background load
background_load = cs.loop_stress(f'mixed ratio\\(write=1,read=1\\) duration=5m cl=QUORUM -pop dist=UNIFORM\\(1..{ROW_COUNT}\\) -log hdrfile=profile.hdr -graph file=report.html title=benchmark revision=benchmark-0 -mode native cql3 maxPending=1024 -rate "threads=700 fixed={BACKGROUND_LOAD_OPS // loadgenerator_count}/s" -node {cluster_string}')

add_nodes_start = datetime.now()

iteration = Iteration(f'{profile_name}/add-node', ignore_git=True)

# Start Scylla/Cassandra nodes
if props['cluster_type'] == 'scylla':
    s = Scylla(new_node_public_ips, new_node_private_ips, all_private_ips[0], props)
    s.start()
else:
    cassandra = Cassandra(new_node_public_ips, new_node_private_ips, all_private_ips[0], props)
    cassandra.start()

add_nodes_end = datetime.now()

print("Run ended at:", datetime.now().strftime("%H:%M:%S"))
print("Adding nodes took:", (add_nodes_end - add_nodes_start).total_seconds(), "seconds.")

with open(f'{iteration.dir}/result.txt', 'a') as writer:
    writer.write(f'Adding nodes took (s): {(add_nodes_end - add_nodes_start).total_seconds()}\n')

background_load.request_stop()
background_load.join()
print("Background load ended:", datetime.now().strftime("%H:%M:%S"))



from datetime import datetime
from scyllaso.ssh import PSSH, SSH
from scyllaso.util import run_parallel, find_java, log_important, log_machine
from scyllaso.network_wait import wait_for_cql_start
from scyllaso.raid import RAID


class Cassandra:

    def __init__(self,
                 cluster_public_ips,
                 cluster_private_ips,
                 seed_private_ip,
                 properties,
                 setup_raid=True,
                 cassandra_version=None):
        self.properties = properties
        self.cluster_public_ips = cluster_public_ips
        self.cluster_private_ips = cluster_private_ips
        self.seed_private_ip = seed_private_ip
        if cassandra_version is not None:
            self.cassandra_version = cassandra_version
        else:
            self.cassandra_version = properties['cassandra_version']

        self.ssh_user = properties['cluster_user']
        self.setup_raid = setup_raid
        # trigger early detection of missing java.
        find_java(properties)

    def __new_ssh(self, ip):
        return SSH(ip, self.ssh_user, self.properties['ssh_options'])

    def __install(self, ip):
        ssh = self.__new_ssh(ip)
        ssh.update()
        log_machine("Installing Cassandra: started", ip)
        ssh.install_one('openjdk-16-jdk', 'java-16-openjdk')
        ssh.install('wget')
        private_ip = self.__find_private_ip(ip)
        path_prefix = 'cassandra-raid/' if self.setup_raid else './'
        ssh.exec(f"""
            set -e
            
            if [ -d '{path_prefix}apache-cassandra-{self.cassandra_version}' ]; then
                echo Cassandra {self.cassandra_version} already installed.
                exit 0
            fi
            
            wget -q -N https://archive.apache.org/dist/cassandra/{self.cassandra_version}/apache-cassandra-{self.cassandra_version}-bin.tar.gz
            tar -xzf apache-cassandra-{self.cassandra_version}-bin.tar.gz -C {path_prefix}
            
            wget -q https://github.com/criteo/cassandra_exporter/releases/download/2.3.5/cassandra_exporter-2.3.5.jar
        """)
        ssh.scp_to_remote("jvm11-server.options",
                          f"{path_prefix}apache-cassandra-{self.cassandra_version}/conf/jvm11-server.options")
        ssh.scp_to_remote("cassandra.yaml",
                          f"{path_prefix}apache-cassandra-{self.cassandra_version}/conf/cassandra.yaml")
        ssh.scp_to_remote("cassandra-exporter.yml", f"config.yml")
        # FIXME - heap in MB * 2
        ssh.exec("""
            sudo sh -c "echo 262144 > /proc/sys/vm/max_map_count"
        """)
        ssh.exec(f"""
            cd {path_prefix}apache-cassandra-{self.cassandra_version}
            sudo sed -i \"s/seeds:.*/seeds: {self.seed_private_ip} /g\" conf/cassandra.yaml
            sudo sed -i \"s/listen_address:.*/listen_address: {private_ip} /g\" conf/cassandra.yaml
            sudo sed -i \"s/rpc_address:.*/rpc_address: {private_ip} /g\" conf/cassandra.yaml
        """)
        log_machine("Installing Cassandra: done", ip)

    def __find_private_ip(self, public_ip):
        index = self.cluster_public_ips.index(public_ip)
        return self.cluster_private_ips[index]

    def nodetool(self, command, load_index=None):
        if load_index is None:
            run_parallel(self.nodetool, [(command, i) for i in range(len(self.cluster_private_ips))])
        else:
            path_prefix = 'cassandra-raid/' if self.setup_raid else './'
            ssh = self.__new_ssh(self.cluster_public_ips[load_index])
            ssh.exec(f"{path_prefix}apache-cassandra-{self.cassandra_version}/bin/nodetool {command}")

    def install(self):
        log_important("Installing Cassandra: started")
        if self.setup_raid:
            log_important("Installing Cassandra: setting up RAID")
            raid = RAID(self.cluster_public_ips, self.ssh_user, '/dev/nvme*n1', 'cassandra-raid', 0, self.properties)
            raid.install()
            log_important("Installing Cassandra: finished setting up RAID")
        run_parallel(self.__install, [(ip,) for ip in self.cluster_public_ips])
        log_important("Installing Cassandra: done")

    def __start(self, ip):
        log_machine("Starting Cassandra: started", ip)
        ssh = self.__new_ssh(ip)
        path_prefix = 'cassandra-raid/' if self.setup_raid else './'
        ssh.exec(f"""
            set -e
            cd {path_prefix}apache-cassandra-{self.cassandra_version}
            if [ -f 'cassandra.pid' ]; then
                pid=$(cat cassandra.pid)
                kill $pid || true
                while kill -0 $pid; do 
                    sleep 1
                done
                rm -f 'cassandra.pid'
            fi
            bin/cassandra -p cassandra.pid 2>&1 >> cassandra.out &
            """)
        log_machine("Starting Cassandra: done", ip)

    def __start_exporter(self, ip):
        print(f'    [{ip}] Starting exporter')
        ssh = self.__new_ssh(ip)
        ssh.exec(f"""
            java -Dorg.slf4j.simpleLogger.defaultLogLevel=trace -jar ~/cassandra_exporter-2.3.5.jar ~/config.yml >> ~/cassandra_exporter.out 2>&1 &
            """)
        print(f'    [{ip}] Starting exporter: done')

    def append_env_configuration(self, configuration):
        print(f"Appending cassandra-env.sh configuration to nodes {self.cluster_public_ips}: {configuration}")
        pssh = PSSH(self.cluster_public_ips, self.ssh_user, self.properties['ssh_options'])
        path_prefix = 'cassandra-raid/' if self.setup_raid else './'
        pssh.exec(
            f"echo '{configuration}' >> {path_prefix}apache-cassandra-{self.cassandra_version}/conf/cassandra-env.sh")
        print(f"echo '{configuration}' >> {path_prefix}apache-cassandra-{self.cassandra_version}/conf/cassandra-env.sh")

    def start(self):
        log_important(f"Starting Cassandra nodes {self.cluster_public_ips}")
        for public_ip in self.cluster_public_ips:
            self.__start(public_ip)
            wait_for_cql_start(public_ip)
            log_important("Node finished bootstrapping at:", datetime.now().strftime("%H:%M:%S"), public_ip)
            self.__start_exporter(public_ip)
        log_important(f"Starting Cassandra nodes {self.cluster_public_ips}: done")

    def __stop(self, ip):
        log_machine(ip, "Stopping Cassandra: started")
        ssh = self.__new_ssh(ip)
        path_prefix = 'cassandra-raid/' if self.setup_raid else './'
        ssh.exec(f"""
            set -e
            cd {path_prefix}apache-cassandra-{self.cassandra_version}
            if [ -f 'cassandra.pid' ]; then
                pid=$(cat cassandra.pid)
                kill $pid || true
                while kill -0 $pid; do 
                    sleep 1
                done
                rm -f 'cassandra.pid'
            fi
            """)

        print(f'    [{ip}] Stopping Cassandra: done')

    def stop(self, load_index=None, erase_data=False):
        if load_index is None:
            log_important("Stop Cassandra: started")
            run_parallel(self.__stop, [(ip,) for ip in self.cluster_public_ips])
            log_important("Stop Cassandra: done")
        else:
            self.__stop(self.cluster_public_ips[load_index])

            if erase_data:
                ssh = self.__new_ssh(self.cluster_public_ips[load_index])
                path_prefix = 'cassandra-raid/' if self.setup_raid else './'
                ssh.exec(f"rm -rf {path_prefix}apache-cassandra-{self.cassandra_version}/data")



#!/bin/python3

import sys
from scyllaso import common
from scyllaso.cs import CassandraStress
from scyllaso.scylla import Scylla
from datetime import datetime

print("Test started at:", datetime.now().strftime("%H:%M:%S"))

if len(sys.argv) < 2:
    raise Exception("Usage: ./benchmark_drivers.py [PROFILE_NAME]")

# Load properties
profile_name = sys.argv[1]
props = common.load_yaml(f'{profile_name}.yml')
env = common.load_yaml(f'environment_{profile_name}.yml')
cluster_private_ips = env['cluster_private_ips']
cluster_string = ",".join(cluster_private_ips)
cluster_public_ips = env['cluster_public_ips']
loadgenerator_public_ips = env['loadgenerator_public_ips']
loadgenerator_count = len(loadgenerator_public_ips)

ROW_SIZE_BYTES = 300

# 20GB per node
TARGET_DATASET_SIZE = len(cluster_private_ips) * 20 * 1024 * 1024 * 1024

REPLICATION_FACTOR = 3
ROW_COUNT = int(TARGET_DATASET_SIZE / ROW_SIZE_BYTES / REPLICATION_FACTOR)

# Start Scylla nodes
s = Scylla(env['cluster_public_ips'], env['cluster_private_ips'], env['cluster_private_ips'][0], props)
s.install()
s.start()

print("Nodes started at:", datetime.now().strftime("%H:%M:%S"))

cs = CassandraStress(env['loadgenerator_public_ips'], props, scylla_tools=False)
cs.install()
cs.prepare()

print("Loading started at:", datetime.now().strftime("%H:%M:%S"))

cs.stress_seq_range(ROW_COUNT, 'write cl=QUORUM', f'-schema "replication(strategy=SimpleStrategy,replication_factor={REPLICATION_FACTOR})" -log hdrfile=profile.hdr -graph file=report.html title=benchmark revision=benchmark-0 -mode native cql3 -rate "threads=300 throttle=50000/s" -node {cluster_string}')

print("Loading ended at:", datetime.now().strftime("%H:%M:%S"))


#!/bin/python3

import sys
import time
from scyllaso import common
from scyllaso.cs import CassandraStress
from scyllaso.common import Iteration
from scyllaso.scylla import Scylla
from scyllaso.cassandra import Cassandra
from datetime import datetime

print("Test started at:", datetime.now().strftime("%H:%M:%S"))

if len(sys.argv) < 2:
    raise Exception("Usage: ./benchmark_repair.py [PROFILE_NAME]")

# Load properties
profile_name = sys.argv[1]
props = common.load_yaml(f'{profile_name}.yml')
env = common.load_yaml(f'environment_{profile_name}.yml')
cluster_private_ips = env['cluster_private_ips']
cluster_string = ",".join(cluster_private_ips)
cluster_public_ips = env['cluster_public_ips']
loadgenerator_public_ips = env['loadgenerator_public_ips']
loadgenerator_count = len(loadgenerator_public_ips)

# Run parameters

# Row size of default cassandra-stress workload.
# Measured experimentally.
ROW_SIZE_BYTES = 210 * 1024 * 1024 * 1024 / 720_000_000

# 1TB per node
TARGET_DATASET_SIZE = props['target_dataset_size_gb'] * 1024 * 1024 * 1024

REPLICATION_FACTOR = 3
COMPACTION_STRATEGY = props['compaction_strategy']
ROW_COUNT = int(TARGET_DATASET_SIZE / ROW_SIZE_BYTES / REPLICATION_FACTOR)

# Start Scylla/Cassandra nodes
if props['cluster_type'] == 'scylla':
    cluster = Scylla(env['cluster_public_ips'], env['cluster_private_ips'], env['cluster_private_ips'][0], props)
    cluster.install()
    cluster.start()
else:
    cluster = Cassandra(env['cluster_public_ips'], env['cluster_private_ips'], env['cluster_private_ips'][0], props)
    cluster.install()
    if "cassandra_extra_env_opts" in props:
        cluster.append_env_configuration(props["cassandra_extra_env_opts"])
    cluster.start()

print("Nodes started at:", datetime.now().strftime("%H:%M:%S"))

cs = CassandraStress(env['loadgenerator_public_ips'], props)
cs.install()
cs.prepare()

print("Loading started at:", datetime.now().strftime("%H:%M:%S"))

cs.stress_seq_range(ROW_COUNT, 'write cl=QUORUM',
                    f'-schema "replication(strategy=SimpleStrategy,replication_factor={REPLICATION_FACTOR})" "compaction(strategy={COMPACTION_STRATEGY})" -log hdrfile=profile.hdr -graph file=report.html title=benchmark revision=benchmark-0 -mode native cql3 -rate "threads=700 throttle=33000/s" -node {cluster_string}')

print("Sleeping 2h")
time.sleep(60 * 60 * 2)

print("Run started at:", datetime.now().strftime("%H:%M:%S"))

iteration = Iteration(f'{profile_name}/repair_no_errors', ignore_git=True)

repair_start = datetime.now()

print("Reparing node 0 started at:", datetime.now().strftime("%H:%M:%S"))
cluster.nodetool("repair -full", 0)
print("Reparing node 0 ended at:", datetime.now().strftime("%H:%M:%S"))

repair_end = datetime.now()

with open(f'{iteration.dir}/result.txt', 'a') as writer:
    writer.write(f'Reparing node 0 took (s): {(repair_end - repair_start).total_seconds()}\n')

print("Run ended at:", datetime.now().strftime("%H:%M:%S"))



import os
from datetime import datetime
from scyllaso.ssh import SSH
from scyllaso.util import run_parallel,log_important,log_machine,log


class Fio:

    def __init__(self, ips, ssh_user, ssh_options, capture_lsblk=True):
        self.ips = ips
        self.ssh_user = ssh_user
        self.ssh_options = ssh_options
        self.dir_name = "fio-" + datetime.now().strftime("%d-%m-%Y_%H-%M-%S")
        self.capture_lsblk = capture_lsblk

    def __new_ssh(self, ip):
        return SSH(ip, self.ssh_user, self.ssh_options)

    def __upload(self, ip, file):
        ssh = self.__new_ssh(ip)
        ssh.exec(f"mkdir -p {self.dir_name}")
        ssh.scp_to_remote(file, self.dir_name)

    def upload(self, file):
        log_important(f"Upload: started")
        run_parallel(self.__upload, [(ip, file) for ip in self.ips])
        log_important(f"Upload-Stress: done")

    def __install(self, ip):
        log_machine(ip, 'Installing fio: started')
        ssh = self.__new_ssh(ip)
        ssh.update()
        ssh.install('fio')
        log_machine(ip, 'Installing fio: done')

    def install(self):
        log_important(f"fio Installation: started")
        run_parallel(self.__install, [(ip,) for ip in self.ips])
        log_important(f"fio Installation: done")

    def __run(self, ip, options):
        log_machine(ip, 'fio: started')
        ssh = self.__new_ssh(ip)
        if self.capture_lsblk:
            ssh.exec(f'lsblk > lsblk.txt')

        ssh.exec(f"""
            mkdir -p {self.dir_name}
            cd {self.dir_name}
            sudo fio {options}            
            """)
        log_machine(ip, 'fio: done')

    def run(self, options):
        log_important(f"fio run: started")
        log(f"sudo fio {options}")
        run_parallel(self.__run, [(ip, options) for ip in self.ips])
        log_important(f"fio run: done")

    def __download(self, ip, dir):
        dest_dir = os.path.join(dir, ip)
        os.makedirs(dest_dir, exist_ok=True)

        log_machine(ip, f'Downloading to [{dest_dir}]')
        ssh = self.__new_ssh(ip)
        ssh.scp_from_remote(f'{self.dir_name}/*', dest_dir)
        if self.capture_lsblk:
            self.__new_ssh(ip).scp_from_remote(f'lsblk.out', dest_dir)

        log_machine(ip, f'Downloading to [{dest_dir}] done')

    def download(self, dir):
        log_important(f"fio download: started")
        run_parallel(self.__download, [(ip, dir) for ip in self.ips])
        log_important(f"fio download: done")





import sys
import os
from scyllaso import util


def cli():
    if len(sys.argv) == 1:
        print("No data directory")
        exit(1)

    old_wd = os.getcwd()
    data_dir = sys.argv[1]

    if not os.path.isabs(data_dir):
        data_dir = os.path.join(old_wd, data_dir)

    if not os.path.exists(data_dir):
        print(f"[{data_dir}] does not exist!")
        exit(1)

    if not os.path.isdir(data_dir):
        print(f"[{data_dir}] is not a directory!")
        exit(1)

    scylla_monitoring_path = os.environ['SCYLLA_MONITORING']

    old_wd = os.getcwd()
    os.chdir(scylla_monitoring_path)

    util.log_important("Killing Scylla Monitoring: starting")
    os.system("./kill-all.sh")
    util.log_important("Killing Scylla Monitoring: done")

    util.log_important("Starting Scylla Monitoring: started")
    os.system(f"./start-all.sh -d {data_dir} -s prometheus/scylla_servers.example.yml")
    util.log_important("Starting Scylla Monitoring: done")

    os.chdir(old_wd)



import os


def cli():
    scylla_monitoring_path = os.environ['SCYLLA_MONITORING']

    old_wd = os.getcwd()
    os.chdir(scylla_monitoring_path)
    cmd = f"./kill-all.sh"
    os.system(cmd)
    os.chdir(old_wd)



#!/bin/python3

from scyllaso import common
from scyllaso import scylla
from scyllaso.cs import CassandraStress
from scyllaso.common import Iteration
from scyllaso import prometheus

# Load the properties
props = common.load_yaml('properties.yml')

# Provision the machines.
# terraform.apply(props['terraform_plan'])

# Load information about the created machines.
env = common.load_yaml('environment.yml')
cluster_private_ips = env['cluster_private_ips']
cluster_string = ",".join(cluster_private_ips)

iteration = Iteration("dummy-benchmark")

# Setup cassandra stress
cs = CassandraStress(env['loadgenerator_public_ips'], props)
cs.install()
cs.prepare()

# Total number of items.
items = 10_000_000

# The running time of the benchmark
duration = "12m"
# 2 minutes of warmup are removed from the duration.
warmup_seconds = 120

# Configuration for throughput test 
rate = f'threads=200'
# Configuration for a latency: fixed number of requests per second
# rate = f'threads=200 fixed="2000/s"'

cs.upload("stress_example.yaml")

# Insert the test data.
cs.insert("stress_example.yaml", items, cluster_string)

# Restart to cluster to make sure the Scylla starts fresh 
# e.g. the memtable is flushed.
scylla.restart_cluster(env['cluster_public_ips'], props['cluster_user'], props['ssh_options'])

# Actual benchmark
cs.stress(f'user profile=./stress_example.yaml "ops(insert=1)" duration={duration} -pop seq=1..{items} -log hdrfile=profile.hdr -graph file=report.html title=benchmark revision=benchmark-0 -mode native cql3 -rate {rate} -node {cluster_string}')

# collect the results.
cs.collect_results(iteration.dir, warmup_seconds=warmup_seconds)

# Download and clear the prometheus data (can take a lot of time/space)
prometheus.download_and_clear(env, props, iteration)

# Automatically terminates the cluster.
#terraform.destroy(props['terraform_plan'])
print("Call 'unprovision_terraform.py' to destroy the created infrastructure!")



import sys
import os
import argparse
from scyllaso import common
from scyllaso.perf import Perf


def cli():
    parser = argparse.ArgumentParser()
    parser.add_argument("CPU", help="The CPU to profile", nargs="?")
    parser.add_argument("-o", "--output", help="The name (no extension) of the output file", default="flamegraph")
    parser.add_argument("-a", "--args", help="The extra perf record arguments", default="--call-graph lbr -F99")
    parser.add_argument("-d", "--duration", help="Duration to profile in seconds (default 30) ", type=int, default=30)
    parser.add_argument("-l", "--list", help="Runs `perf list -v`", action='store_true')
    args = parser.parse_args()

    props = common.load_yaml('properties.yml')
    env = common.load_yaml('environment.yml')
    cluster_public_ips = env['cluster_public_ips']

    perf = Perf(cluster_public_ips, props['cluster_user'], props['ssh_options'])
    perf.install()

    if args.list:
        perf.list()
    else:
        cpu = args.CPU
        if not cpu:
            print("No CPU was selected")
            sys.exit(1)

        print(f"Profiling CPU {cpu}")
        perf.flamegraph_cpu(cpu, os.getcwd(), duration_seconds=args.duration, args=args.args, output=args.output)



from setuptools import setup, find_packages

setup(
    name='scyllaso',
    version='0.1.dev2',
    author='Peter Veentjer',
    author_email='peter.veentjer@scylladb.com',
    description='The Scylla Stress Orchestrator is Python 3 based framework for running various '
                'benchmark tools including cassandra-stress and scylla-bench.',
    long_description='The Scylla Stress Orchestrator is Python 3 based framework for running various '
                'benchmark tools including cassandra-stress and scylla-bench.',
    long_description_content_type='text/markdown',
    url='https://github.com/scylladb/scylla-stress-orchestrator',
    packages=find_packages(),
    python_requires='>=3.7',
    project_urls={
        'Bug Tracker': 'https://github.com/scylladb/scylla-stress-orchestrator/issues',
    },
    license_files=('LICENSE.txt',),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: POSIX :: Linux',
    ],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'make_key = scyllaso.bin.make_key:cli',
            'scylla_monitoring_start = scyllaso.bin.scylla_monitoring_start:cli',
            'scylla_monitoring_stop = scyllaso.bin.scylla_monitoring_stop:cli',
            'make_cpu_config = scyllaso.bin.make_cpu_config:cli',
            'kill_load_generators = scyllaso.bin.kill_load_generators:cli',
            'generate_benchmark = scyllaso.bin.generate_benchmark:cli',
            'flamegraph_cpu = scyllaso.bin.flamegraph_cpu:cli',
            'provision_terraform = scyllaso.bin.provision_terraform:provision',
            'unprovision_terraform = scyllaso.bin.provision_terraform:unprovision',
        ],
    }
)



import os
import time

from datetime import datetime
from scyllaso.hdr import HdrLogProcessor
from scyllaso.ssh import SSH
from scyllaso.util import run_parallel, WorkerThread, log_important, log_machine, log, WorkerThreadLoop


class CassandraStress:

    def __init__(self, load_ips, properties, scylla_tools=True, performance_governor=True):
        self.properties = properties
        self.load_ips = load_ips
        self.ssh_user = properties.get('load_generator_user')
        self.scylla_tools = scylla_tools
        self.performance_governor = performance_governor

    def __new_ssh(self, ip):
        return SSH(ip, self.ssh_user, self.properties['ssh_options'])

    def __install(self, ip):
        ssh = self.__new_ssh(ip)
        ssh.update()

        if self.performance_governor:
            ssh.set_governor("performance")

        if self.scylla_tools:
            log_machine(ip, f'Installing cassandra-stress (Scylla): started')
            ssh.exec(f"""
                set -e
                if hash apt-get 2>/dev/null; then
                    sudo apt-get install -y apt-transport-https
                    sudo apt-key adv --keyserver hkp://keyserver.ubuntu.com:80 --recv-keys 5e08fbd8b5d6ec9c
                    sudo curl -L --output /etc/apt/sources.list.d/scylla.list http://downloads.scylladb.com/deb/ubuntu/scylla-4.3-$(lsb_release -s -c).list
                    sudo apt-get update -y
                    sudo apt-get install -y scylla-tools
                elif hash yum 2>/dev/null; then
                    sudo yum install  -y -q https://dl.fedoraproject.org/pub/epel/epel-release-latest-7.noarch.rpm
                    sudo curl -o /etc/yum.repos.d/scylla.repo -L http://repositories.scylladb.com/scylla/repo/603fc559-4518-4f8e-8ceb-2851dec4ab23/centos/scylladb-4.3.repo
                    sudo yum install -y -q scylla-tools
                else
                    echo "Cannot install scylla-tools: yum/apt not found"
                    exit 1
                fi
                """)
        else:
            log_machine(ip, f'Installing cassandra-stress (Cassandra): started')
            cassandra_version = self.properties['cassandra_version']
            ssh.install_one('openjdk-8-jdk', 'java-1.8.0-openjdk')
            ssh.install('wget')
            ssh.exec(f"""
                set -e
                wget -q -N https://mirrors.netix.net/apache/cassandra/{cassandra_version}/apache-cassandra-{cassandra_version}-bin.tar.gz
                tar -xzf apache-cassandra-{cassandra_version}-bin.tar.gz
            """)

        log_machine(ip, f'Installing cassandra-stress: done')

    def install(self):
        log_important("Installing Cassandra-Stress: started")
        run_parallel(self.__install, [(ip,) for ip in self.load_ips])
        log_important("Installing Cassandra-Stress: done")

    def __stress(self, ip, startup_delay, cmd):
        time.sleep(startup_delay)

        if self.scylla_tools:
            full_cmd = f'cassandra-stress {cmd}'
        else:
            cassandra_version = self.properties['cassandra_version']
            cassandra_stress_dir = f'apache-cassandra-{cassandra_version}/tools/bin'
            full_cmd = f'{cassandra_stress_dir}/cassandra-stress {cmd}'

        dt = datetime.now().strftime("%d-%m-%Y_%H-%M-%S")
        full_cmd = full_cmd + f" 2>&1 | tee -a cassandra-stress-{dt}.log"
        log(full_cmd)
        self.__new_ssh(ip).exec(full_cmd)

    def stress(self, command, load_index=None):
        if load_index is None:
            log_important("Cassandra-Stress: started")
            run_parallel(self.__stress, [(ip, 10 if i > 0 else 0, command) for i, ip in enumerate(self.load_ips)])
            log_important("Cassandra-Stress: done")
        else:
            log("using load_index " + str(load_index))
            self.__stress(self.load_ips[load_index], 0, command)

    def stress_seq_range(self, row_count, command_part1, command_part2):
        load_ip_count = len(self.load_ips)
        row_count_per_ip = row_count // load_ip_count
        range_points = [1]
        for i in range(load_ip_count):
            range_points.append(range_points[-1] + row_count_per_ip)
        range_points[-1] = row_count

        population_commands = []
        # FIXME - cleanup
        for i in range(len(range_points) - 1):
            population_commands.append(f' n={range_points[i + 1] - range_points[i] + 1} -pop seq={range_points[i]}..{range_points[i + 1]} ')

        print(population_commands)

        log_important("Cassandra-Stress: started")
        run_parallel(self.__stress, [(ip, 10 if i > 0 else 0, command_part1 + pop_command + command_part2) for i, (ip, pop_command) in enumerate(zip(self.load_ips, population_commands))])
        log_important("Cassandra-Stress: done")

    def async_stress(self, command, load_index=None):
        thread = WorkerThread(self.stress, (command, load_index))
        thread.start()
        return thread.future

    def loop_stress(self, command, load_index=None):
        thread = WorkerThreadLoop(self.stress, (command, load_index))
        thread.start()
        return thread

    def insert(self, profile, item_count, nodes, mode="native cql3", rate="threads=100", sequence_start=None):
        log_important(f"Inserting {item_count} items")
        start_seconds = time.time()

        per_load_generator = item_count // len(self.load_ips)
        start = sequence_start
        if sequence_start is None:
            start = 1

        end = start + per_load_generator - 1
        cmd_list = []
        for i in range(0, len(self.load_ips)):
            cmd = f'user profile={profile} "ops(insert=1)" n={per_load_generator} no-warmup -pop seq={start}..{end} -mode {mode} -rate {rate}  -node {nodes}'
            log(self.load_ips[i] + " " + cmd)
            cmd_list.append(cmd)
            start = end + 1
            end = end + per_load_generator

        futures = []
        for i in range(0, len(self.load_ips)):
            f = self.async_stress(cmd_list[i], load_index=i)
            futures.append(f)
            if i == 0:
                time.sleep(10)

        for f in futures:
            f.join()

        duration_seconds = time.time() - start_seconds
        log(f"Duration : {duration_seconds} seconds")
        log(f"Insertion rate: {item_count // duration_seconds} items/second")
        log_important(f"Inserting {item_count} items: done")

    def __ssh(self, ip, command):
        self.__new_ssh(ip).exec(command)

    def ssh(self, command):
        run_parallel(self.__ssh, [(ip, command) for ip in self.load_ips])

    def __upload(self, ip, file):
        self.__new_ssh(ip).scp_to_remote(file, "")

    def upload(self, file):
        log_important(f"Upload: started")
        run_parallel(self.__upload, [(ip, file) for ip in self.load_ips])
        log_important(f"Upload: done")

    def __collect(self, ip, dir):
        dest_dir = os.path.join(dir, ip)
        os.makedirs(dest_dir, exist_ok=True)
        log_machine(ip, f'Collecting to [{dest_dir}]')
        ssh = self.__new_ssh(ip)
        ssh.scp_from_remote(f'*.{{html,hdr,log}}', dest_dir)
        ssh.exec(f'rm -fr *.html *.hdr *.log')
        log_machine(ip, f'Collecting to [{dest_dir}] done')

    def collect_results(self, dir, warmup_seconds=None, cooldown_seconds=None):
        """
        Parameters
        ----------
        dir: str
            The download directory.
        warmup_seconds : str
            The warmup period in seconds. If the value is set, additional files will 
            be created where the warmup period is trimmed.
        cooldown_seconds : str
            The cooldown period in seconds. If the value is set, additional files will 
            be created where the cooldown period is trimmed.            
        """

        log_important(f"Collecting results: started")
        run_parallel(self.__collect, [(ip, dir) for ip in self.load_ips])
        p = HdrLogProcessor(self.properties, warmup_seconds=warmup_seconds, cooldown_seconds=cooldown_seconds)
        p.trim_recursivly(dir)
        p.merge_recursivly(dir)
        p.process_recursivly(dir)
        p.summarize_recursivly(dir)
        log_important(f"Collecting results: done")
        log(f"Results can be found in [{dir}]")

    def __prepare(self, ip):
        log_machine(ip, f'Preparing: started')
        ssh = self.__new_ssh(ip)
        # we need to make sure that the no old load generator is still running.
        ssh.exec(f'killall -q -9 java')
        log_machine(ip, f'Preparing: done')

    def prepare(self):
        log_important(f"Preparing load generator: started")
        run_parallel(self.__prepare, [(ip,) for ip in self.load_ips])
        log_important(f"Preparing load generator: done")



import enum
import shlex
import subprocess
import selectors
from datetime import datetime
from threading import Thread
from threading import Lock, Condition


class Future:
    def __init__(self):
        self.__condition = Condition(Lock())
        self.__val = None
        self.__is_set = False

    def get(self):
        with self.__condition:
            while not self.__is_set:
                self.__condition.wait()

            if self.__val is Exception:
                raise Exception() from self.__val
            return self.__val

    def join(self):
        self.get()

    def set(self, val):
        with self.__condition:
            if self.__is_set:
                raise RuntimeError("Future has already been set")
            self.__val = val
            self.__is_set = True
            self.__condition.notify_all()

    def done(self):
        return self.__is_set


class WorkerThread(Thread):

    def __init__(self, target, args):
        super().__init__(target=target, args=args)
        self.future = Future()

    def run(self):
        self.exception = None
        try:
            super().run()
            self.future.set(True)
        except Exception as e:
            self.exception = e
            self.future.set(e)


def run_parallel(target, args_list, ignore_errors=False):
    # print(f"parallel{type(args_list)}")
    # print(args_list)
    threads = []
    for args in args_list:
        # print(f"run_parallel loop |{args}|")
        # print(f"type {type(args)}")
        thread = WorkerThread(target, args)
        thread.start()
        threads.append(thread)
    for thread in threads:
        thread.join()
        if not ignore_errors and thread.exception:
            raise Exception() from thread.exception


class WorkerThreadLoop(Thread):

    def __init__(self, target, args):
        super().__init__(target=target, args=args)
        self.stopped = False

    def run(self):
        self.exception = None
        while not self.stopped:
            try:
                if self._target is not None:
                    self._target(*self._args, **self._kwargs)
            except Exception as e:
                self.exception = e
                break

    def request_stop(self):
        self.stopped = True


def find_java(properties):
    path = properties.get("jvm_path")
    if path:
        return f"{path}/bin/java"
    from shutil import which
    path = which("java")
    if path:
        return path
    else:
        raise RuntimeError("Could not locate java")


def join_all(*futures):
    for f in futures:
        f.join()


def call(cmd, shell=False, split=True):
    if split:
        cmd = cmd.split()
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=shell)

    sel = selectors.DefaultSelector()
    sel.register(process.stdout, selectors.EVENT_READ)
    sel.register(process.stderr, selectors.EVENT_READ)

    while True:
        for key, _ in sel.select():
            data = key.fileobj.read1().decode()

            if not data:
                return process.poll()
            lines = data.splitlines()
            log_level = LogLevel.info if key.fileobj is process.stdout else LogLevel.warning
            for line in lines:
                log(line, log_level)


class LogLevel(enum.Enum):
    info = 1
    warning = 2
    error = 3


def log_machine(ip, text, log_level=LogLevel.info):
    if not text:
        return

    prefix = f"{ip}".ljust(15, " ") + "    "
    log(f"{prefix} {text}", log_level=log_level)


def level_text(log_level):
    if log_level == LogLevel.info:
        return "INFO "
    elif log_level == LogLevel.warning:
        return "WARN "
    else:
        return "ERROR"


def log(text, log_level=LogLevel.info):
    if not text:
        return

    dt = datetime.now().strftime("%H:%M:%S")
    level_txt = level_text(log_level)
    print(f"{dt} {level_txt} {text}")


def log_important(text):
    l = 80 - len(text)
    if l > 0:
        s = '-' * l
    else:
        s = ''
    log(f"-------------[ {text} ]{s}-----")



import os
import time

from scyllaso import common

def env_to_inventory():
    props = common.load_yaml('properties.yml')
    cluster_user = props['cluster_user']
    prometheus_user = props['prometheus_user']

    env = common.load_yaml('environment.yml')
    cluster_private_ips = env['cluster_private_ips']
    cluster_public_ips = env['cluster_public_ips']

    key = "/eng/scylla/scylla-ansible-roles/ansible-scylla-node/foo/key"
    
    inventory = "#inventory.ini\n"
    inventory = "[scylla-manager]\n"
    prometheus_public_ip = env['prometheus_public_ip']
    inventory = f"{prometheus_public_ip} ansible_connection=ssh ansible_user={prometheus_user} ansible_ssh_private_key_file={key}\n\n"

    inventory += "[scylla]\n"
    extra = f"ansible_connection=ssh ansible_user={cluster_user} ansible_ssh_private_key_file={key} dc=us-east1 rack=rack-b"
    
    for public_ip in cluster_public_ips:
        line = f"{public_ip} {extra}\n"
        inventory += line

    inventory_file = open('inventory.ini', 'w')
    inventory_file.write(inventory)
    inventory_file.close()    



from scyllaso.ssh import SSH
from scyllaso.util import log_important


def download(env, props, iteration):
    prometheus = Prometheus(env['prometheus_public_ip'][0],
                            props['prometheus_user'],
                            props['ssh_options'])
    prometheus.stop()
    prometheus.data_dir_download(iteration.dir)
    prometheus.start()


def download_and_clear(env, props, iteration):
    prometheus = Prometheus(env['prometheus_public_ip'][0],
                            props['prometheus_user'],
                            props['ssh_options'])
    prometheus.stop()
    prometheus.data_dir_download(iteration.dir)
    prometheus.data_dir_rm()
    prometheus.start()


class Prometheus:

    def __init__(self, ip, user, ssh_options, scylla_version="4.4"):
        self.ip = ip
        self.user = user
        self.ssh_options = ssh_options
        self.scylla_version = scylla_version

    def data_dir_upload(self, dir):
        log_important("Prometheus upload: started")
        ssh = SSH(self.ip, self.user, self.ssh_options)
        ssh.scp_to_remote(dir + "/*", "data")
        log_important("Prometheus upload: done")

    def stop(self):
        log_important("Prometheus stop: started")
        ssh = SSH(self.ip, self.user, self.ssh_options)
        ssh.exec(
            f"""
            dir=$(find . -maxdepth 1 -type d -name "scylla-monitoring*" -print -quit)   
            echo "directory [$dir]"         
            cd $dir
            ./kill-all.sh
            """)
        log_important("Prometheus stop: done")

    def start(self):
        log_important("Prometheus start: started")
        ssh = SSH(self.ip, self.user, self.ssh_options)
        ssh.exec(
            f"""
            mkdir -p data
            dir=$(find . -maxdepth 1 -type d -name "scylla-monitoring*" -print -quit)
            cd $dir            
            ./start-all.sh -v {self.scylla_version} -d ../data
            """)
        log_important("Prometheus start: done")

    def data_dir_download(self, dir):
        log_important("Prometheus download data: started")
        ssh = SSH(self.ip, self.user, self.ssh_options)
        ssh.scp_from_remote(f"data", dir)
        log_important("Prometheus download data: done")

    def data_dir_rm(self):
        log_important("Prometheus clear data: started")
        ssh = SSH(self.ip, self.user, self.ssh_options)
        ssh.exec("rm -fr data")
        log_important("Prometheus clear data: done")



import os
import glob
import csv
import pkg_resources
from collections import namedtuple
from scyllaso.util import find_java, log_important, log


class HdrLogProcessor:

    def __init__(self, properties, warmup_seconds=None, cooldown_seconds=None):
        self.properties = properties
        self.java_path = find_java(self.properties)
        self.warmup_seconds = warmup_seconds
        self.cooldown_seconds = cooldown_seconds
        module_dir = os.path.dirname(pkg_resources.resource_filename('scyllaso', '__init__.py'))
        self.lib_dir = os.path.join(module_dir, "lib")
        print("lib_dir:" + str(self.lib_dir))

    def __trim(self, file):
        filename = os.path.basename(file)
        filename_no_ext = os.path.splitext(filename)[0]

        old_cwd = os.getcwd()
        new_cwd = os.path.dirname(os.path.realpath(file))
        os.chdir(new_cwd)

        args = f'union -if {filename} -of trimmed_{filename_no_ext}.hdr'
        if self.warmup_seconds is not None:
            args = f'{args} -start {self.warmup_seconds}'
        if self.cooldown_seconds is not None:
            args = f'{args} -end {self.cooldown_seconds}'

        cmd = f'{self.java_path} -cp {self.lib_dir}/processor.jar CommandDispatcherMain {args}'
        log(cmd)
        os.system(cmd)
        os.chdir(old_cwd)

    def trim_recursivly(self, dir):
        if self.warmup_seconds is None and self.warmup_seconds is None:
            return

        log_important("HdrLogProcessor.trim_recursively")

        for hdr_file in glob.iglob(dir + '/*/*.hdr', recursive=True):
            filename = os.path.basename(hdr_file)
            if filename.startswith("trimmed_"):
                continue

            log(hdr_file)
            self.__trim(hdr_file)

        log_important("HdrLogProcessor.trim_recursively")

    def merge_recursivly(self, dir):
        log_important("HdrLogProcessor.merge_recursively")
        log(dir)
        # todo be careful with merging the merge file.
        files_map = {}

        for hdr_file in glob.iglob(dir + '/*/*.hdr', recursive=True):
            log(hdr_file)
            base = os.path.splitext(os.path.basename(hdr_file))[0]
            files = files_map.get(base)
            if files is None:
                files = []
                files_map[base] = files
            files.append(hdr_file)

        for name, files in files_map.items():
            input = ""
            for file in files:
                input = input + " -ifp " + file
            cmd = f'{self.java_path} -cp {self.lib_dir}/processor.jar CommandDispatcherMain union {input} -of {dir}/{name}.hdr'
            log(cmd)
            os.system(cmd)

        log_important("HdrLogProcessor.merge_recursively")

    def __summarize(self, file):
        filename = os.path.basename(file)
        filename_no_ext = os.path.splitext(filename)[0]
        old_cwd = os.getcwd()
        new_cwd = os.path.dirname(os.path.realpath(file))
        os.chdir(new_cwd)

        summary_text_name = f"{filename_no_ext}-summary.txt"
        summary_csv_name = f"{filename_no_ext}-summary.csv"

        args = f'-if {filename_no_ext}.hdr'
        os.system(
            f'{self.java_path} -cp {self.lib_dir}/processor.jar CommandDispatcherMain summarize {args} >  {summary_text_name}')

        entries = {}
        with open(summary_text_name, 'r') as summary_text_file:
            for line in summary_text_file:
                row = line.split('=')
                entries[row[0].strip()] = row[1].strip()

        with open(summary_csv_name, 'w') as summary_csv_file:
            header = ','.join(entries.keys())
            content = ','.join(entries.values())
            summary_csv_file.write(f'{header}\n')
            summary_csv_file.write(f'{content}\n')

        os.chdir(old_cwd)

    def summarize_recursivly(self, dir):
        log_important("HdrLogProcessor.summarize_recursively")
        for hdr_file in glob.iglob(dir + '/**/*.hdr', recursive=True):
            log(hdr_file)
            self.__summarize(hdr_file)
        log_important("HdrLogProcessor.summarize_recursively")

    def __process(self, file):
        filename = os.path.basename(file)
        filename_no_ext = os.path.splitext(filename)[0]
        old_cwd = os.getcwd()
        new_cwd = os.path.dirname(os.path.realpath(file))
        os.chdir(new_cwd)

        tags = set()
        with open(filename, "r") as hdr_file:
            reader = csv.reader(hdr_file, delimiter=',')
            # Skip headers
            for i in range(5):
                next(reader, None)
            for row in reader:
                first_column = row[0]
                tag = first_column[4:]
                tags.add(tag)

        for tag in tags:
            # process twice; once to get the csv formatted output and again for the non csv output.
            logprocessor = f'{self.java_path} -cp {self.lib_dir}/HdrHistogram-2.1.9.jar org.HdrHistogram.HistogramLogProcessor'
            args = f'-i {filename} -o {filename_no_ext + "_" + tag} -tag {tag} -csv'
            os.system(f'{logprocessor} {args}')
            os.rename(f'{filename_no_ext + "_" + tag}.hgrm', f'{filename_no_ext + "_" + tag}.hgrm.csv')

            args = f'-i {filename} -o {filename_no_ext + "_" + tag} -tag {tag}'
            os.system(f'{logprocessor} {args}')

        os.chdir(old_cwd)

    def process_recursivly(self, dir):
        log_important("HdrLogProcessor.summarize_recursively")
        for hdr_file in glob.iglob(dir + '/**/*.hdr', recursive=True):
            log(hdr_file)
            self.__process(hdr_file)
        log_important("HdrLogProcessor.summarize_recursively")


ProfileSummaryResult = namedtuple('ProfileSummaryResult',
                                  ['ops_count', 'stress_time_s', 'throughput_per_second', 'mean_latency_ms',
                                   'median_latency_ms', 'p90_latency_ms', 'p99_latency_ms', 'p99_9_latency_ms',
                                   'p99_99_latency_ms', 'p99_999_latency_ms'])


def parse_profile_summary_file(path, operation_name='insert'):
    with open(path) as f:
        config = f.readlines()
        config = [x.strip() for x in config]
        config = dict([x.split('=') for x in config if x])

        ops_count = int(config[f'{operation_name}-rt.TotalCount'])
        stress_time_s = float(config[f'{operation_name}-rt.Period(ms)'].replace(',', '.')) / 1000
        throughput_per_second = float(config[f'{operation_name}-rt.Throughput(ops/sec)'].replace(',', '.'))
        mean_latency_ms = float(config[f'{operation_name}-rt.Mean'].replace(',', '.')) / 1_000_000
        median_latency_ms = float(config[f'{operation_name}-rt.50.000ptile'].replace(',', '.')) / 1_000_000
        p90_latency_ms = float(config[f'{operation_name}-rt.90.000ptile'].replace(',', '.')) / 1_000_000
        p99_latency_ms = float(config[f'{operation_name}-rt.99.000ptile'].replace(',', '.')) / 1_000_000
        p99_9_latency_ms = float(config[f'{operation_name}-rt.99.900ptile'].replace(',', '.')) / 1_000_000
        p99_99_latency_ms = float(config[f'{operation_name}-rt.99.990ptile'].replace(',', '.')) / 1_000_000
        p99_999_latency_ms = float(config[f'{operation_name}-rt.99.999ptile'].replace(',', '.')) / 1_000_000

        return ProfileSummaryResult(
            ops_count=ops_count,
            stress_time_s=stress_time_s,
            throughput_per_second=throughput_per_second,
            mean_latency_ms=mean_latency_ms,
            median_latency_ms=median_latency_ms,
            p90_latency_ms=p90_latency_ms,
            p99_latency_ms=p99_latency_ms,
            p99_9_latency_ms=p99_9_latency_ms,
            p99_99_latency_ms=p99_99_latency_ms,
            p99_999_latency_ms=p99_999_latency_ms)



#!/bin/python3

from scyllaso import common
from scyllaso import scylla
from scyllaso.scylla_bench import ScyllaBench
from scyllaso.common import Iteration
from scyllaso import prometheus

props = common.load_yaml('properties.yml')

env = common.load_yaml('environment.yml')
cluster_private_ips = env['cluster_private_ips']
cluster_string = ",".join(cluster_private_ips)

iteration = Iteration("dummy-benchmark")

bench = ScyllaBench(env['loadgenerator_public_ips'], props)
bench.install()
bench.prepare()

# clear and restarts the cluster. So any config changes will also be picked up
scylla.clear_cluster(env['cluster_public_ips'], props['cluster_user'], props['ssh_options'])

rows = 10_000_000

duration = "2m"

# Insert the test data.
bench.insert(rows, cluster_string)

# Restart to cluster to make sure the Scylla starts fresh 
# e.g. the memtable is flushed.
scylla.restart_cluster(env['cluster_public_ips'], props['cluster_user'], props['ssh_options'])

# Actual benchmark
bench.stress(f'-workload sequential -mode write -partition-count {rows} -nodes {cluster_string} -duration {duration} -host-selection-policy=token-aware')

# collect the results.
bench.collect_results(iteration.dir)

# Download and clear the prometheus data (can take a lot of time/space)
prometheus.download_and_clear(env, props, iteration)

# Automatically terminates the cluster.
#terraform.destroy(props['terraform_plan'])
print("Call 'unprovision_terraform.py' to destroy the created infrastructure!")




from time import sleep
from datetime import datetime, timedelta
import socket


def wait_for_cql_start(node_ip, timeout=7200, connect_timeout=10, max_tries_per_second=2):
    print(f'    [{node_ip}] Waiting for CQL port to start (meaning node bootstrap finished). This could take a while.')

    backoff_interval = 1.0 / max_tries_per_second
    timeout_point = datetime.now() + timedelta(seconds=timeout)

    feedback_interval = 20
    print_feedback_point = datetime.now() + timedelta(seconds=feedback_interval)

    while datetime.now() < timeout_point:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            sock.settimeout(connect_timeout)
            try:
                sock.connect((node_ip, 9042))
            except:
                # There was a problem connecting to CQL port.
                sleep(backoff_interval)
                if datetime.now() > print_feedback_point:
                    print_feedback_point = datetime.now() + timedelta(seconds=feedback_interval)
                    print(f'    [{node_ip}] Still waiting for CQL port to start...')

            else:
                print(f'    [{node_ip}] Successfully connected to CQL port.')
                return

    raise Exception(f'Waiting for CQL to start timed out after {timeout} seconds for node: {node_ip}.')



import os
from datetime import datetime
from scyllaso.ssh import SSH
from scyllaso.util import run_parallel,log_important


class DiskExplorer:

    def __init__(self, ips, ssh_user, ssh_options, capture_lsblk=True):
        self.ips = ips
        self.ssh_user = ssh_user
        self.ssh_options = ssh_options
        self.capture_lsblk = capture_lsblk

    def __new_ssh(self, ip):
        return SSH(ip, self.ssh_user, self.ssh_options)

    def __install(self, ip):
        print(f'    [{ip}] Installing disk-explorer: started')
        ssh = self.__new_ssh(ip)
        ssh.update()
        ssh.install('git', 'fio', 'python3', 'python3-pip')
        ssh.exec(f"""
            sudo pip3 install -qqq matplotlib
            rm -fr diskplorer
            git clone -q https://github.com/scylladb/diskplorer.git
            """)
        print(f'    [{ip}] Installing disk-explorer: done')

    def install(self):
        log_important("Disk Explorer Installation: started")
        run_parallel(self.__install, [(ip,) for ip in self.ips])
        log_important("Disk Explorer Installation: done")

    def __run(self, ip, cmd):
        print(f'    [{ip}] Run: started')
        ssh = self.__new_ssh(ip)
        ssh.exec('rm -fr diskplorer/*.svg')
        ssh.exec(f'rm -fr diskplorer/fiotest.tmp')
        if self.capture_lsblk:
            ssh.exec(f'lsblk > lsblk.out')
        ssh.exec(f"""
            ulimit -n 64000            
            ulimit -n
            ulimit -Sn 64000
            ulimit -Sn
            cd diskplorer                 
            python3 diskplorer.py {cmd}
            """)
        # the file is 100 GB; so we want to remove it.
        ssh.exec(f'rm -fr diskplorer/fiotest.tmp')
        print(f'    [{ip}] Run: done')

    def run(self, command):
       log_important(f'Disk Explorer run: started [{datetime.now().strftime("%H:%M:%S")}]')        
        print(f"python3 diskplorer.py {command}")
        run_parallel(self.__run, [(ip, command) for ip in self.ips])
        log_important(f'Disk Explorer run: done [{datetime.now().strftime("%H:%M:%S")}]')

    def __download(self, ip, dir):
        dest_dir = os.path.join(dir, ip)
        os.makedirs(dest_dir, exist_ok=True)

        print(f'    [{ip}] Downloading to [{dest_dir}]')
        self.__new_ssh(ip).scp_from_remote(f'diskplorer/*.{{svg,csv}}', dest_dir)
        if self.capture_lsblk:
            self.__new_ssh(ip).scp_from_remote(f'lsblk.out', dest_dir)
        print(f'    [{ip}] Downloading to [{dest_dir}] done')

    def download(self, dir):
        log_important("Disk Explorer Download: started")
        run_parallel(self.__download, [(ip, dir) for ip in self.ips])
        log_important("Disk Explorer Download: done")
        print(f"Results can be found in [{dir}]")
     
